@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -28,27 +28,24 @@
  * - struct bftw_queue: The queue of bftw_file's left to explore.  Implemented
  *   as a simple circular buffer.
  *
- * - struct bftw_reader: A reader object that simplifies reading directories and
- *   reporting errors.
- *
  * - struct bftw_state: Represents the current state of the traversal, allowing
  *   various helper functions to take fewer parameters.
  */
 
 #include "bftw.h"
+#include "dir.h"
 #include "dstring.h"
+#include "mtab.h"
 #include "stat.h"
 #include "trie.h"
 #include "util.h"
 #include <assert.h>
-#include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <limits.h>
 #include <stdbool.h>
-#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/stat.h>
 #include <unistd.h>
 
 /**
@@ -73,7 +70,7 @@ struct bftw_file {
 	int fd;
 
 	/** This file's type, if known. */
-	enum bftw_type type;
+	enum bfs_type type;
 	/** The device number, for cycle detection. */
 	dev_t dev;
 	/** The inode number, for cycle detection. */
@@ -231,7 +228,6 @@ static void bftw_cache_add(struct bftw_cache *cache, struct bftw_file *file) {
 /** Remove a bftw_file from the cache. */
 static void bftw_cache_remove(struct bftw_cache *cache, struct bftw_file *file) {
 	assert(cache->size > 0);
-	assert(file->fd >= 0);
 
 	size_t size = --cache->size;
 	size_t i = file->heap_index;
@@ -325,7 +321,7 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	file->refcount = 1;
 	file->fd = -1;
 
-	file->type = BFTW_UNKNOWN;
+	file->type = BFS_UNKNOWN;
 	file->dev = -1;
 	file->ino = -1;
 
@@ -335,32 +331,6 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	return file;
 }
 
-/**
- * Get the appropriate (fd, path) pair for the *at() family of functions.
- *
- * @param file
- *         The file being accessed.
- * @param[out] at_fd
- *         Will hold the appropriate file descriptor to use.
- * @param[in,out] at_path
- *         Will hold the appropriate path to use.
- * @return The closest open ancestor file.
- */
-static struct bftw_file *bftw_file_base(struct bftw_file *file, int *at_fd, const char **at_path) {
-	struct bftw_file *base = file;
-
-	do {
-		base = base->parent;
-	} while (base && base->fd < 0);
-
-	if (base) {
-		*at_fd = base->fd;
-		*at_path += bftw_child_nameoff(base);
-	}
-
-	return base;
-}
-
 /**
  * Open a bftw_file relative to another one.
  *
@@ -377,15 +347,21 @@ static struct bftw_file *bftw_file_base(struct bftw_file *file, int *at_fd, cons
  * @return
  *         The opened file descriptor, or negative on error.
  */
-static int bftw_file_openat(struct bftw_cache *cache, struct bftw_file *file, const struct bftw_file *base, int at_fd, const char *at_path) {
+static int bftw_file_openat(struct bftw_cache *cache, struct bftw_file *file, const struct bftw_file *base, const char *at_path) {
 	assert(file->fd < 0);
 
+	int at_fd = AT_FDCWD;
+	if (base) {
+		at_fd = base->fd;
+		assert(at_fd >= 0);
+	}
+
 	int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
 	int fd = openat(at_fd, at_path, flags);
 
 	if (fd < 0 && errno == EMFILE) {
 		if (bftw_cache_shrink(cache, base) == 0) {
-			fd = openat(base->fd, at_path, flags);
+			fd = openat(at_fd, at_path, flags);
 		}
 	}
 
@@ -414,51 +390,49 @@ static int bftw_file_openat(struct bftw_cache *cache, struct bftw_file *file, co
  *         The opened file descriptor, or negative on error.
  */
 static int bftw_file_open(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
-	int at_fd = AT_FDCWD;
-	const char *at_path = path;
-	struct bftw_file *base = bftw_file_base(file, &at_fd, &at_path);
+	// Find the nearest open ancestor
+	struct bftw_file *base = file;
+	do {
+		base = base->parent;
+	} while (base && base->fd < 0);
 
-	int fd = bftw_file_openat(cache, file, base, at_fd, at_path);
+	const char *at_path = path;
+	if (base) {
+		at_path += bftw_child_nameoff(base);
+	}
+
+	int fd = bftw_file_openat(cache, file, base, at_path);
 	if (fd >= 0 || errno != ENAMETOOLONG) {
 		return fd;
 	}
 
 	// Handle ENAMETOOLONG by manually traversing the path component-by-component
 
-	// -1 to include the root, which has depth == 0
-	size_t offset = base ? base->depth : (size_t)-1;
-	size_t levels = file->depth - offset;
-	if (levels < 2) {
-		return fd;
+	// Use the ->next linked list to temporarily hold the reversed parent
+	// chain between base and file
+	struct bftw_file *cur;
+	for (cur = file; cur->parent != base; cur = cur->parent) {
+		cur->parent->next = cur;
 	}
 
-	struct bftw_file **parents = malloc(levels * sizeof(*parents));
-	if (!parents) {
-		return fd;
-	}
-
-	struct bftw_file *parent = file;
-	for (size_t i = levels; i-- > 0;) {
-		parents[i] = parent;
-		parent = parent->parent;
-	}
-
-	for (size_t i = 0; i < levels; ++i) {
-		fd = bftw_file_openat(cache, parents[i], base, at_fd, parents[i]->name);
-		if (fd < 0) {
+	// Open the files in the chain one by one
+	for (base = cur; base; base = base->next) {
+		fd = bftw_file_openat(cache, base, base->parent, base->name);
+		if (fd < 0 || base == file) {
 			break;
 		}
-
-		base = parents[i];
-		at_fd = fd;
 	}
 
-	free(parents);
+	// Clear out the linked list
+	for (struct bftw_file *next = cur->next; cur != file; cur = next, next = next->next) {
+		cur->next = NULL;
+	}
+
 	return fd;
 }
 
 /**
- * Open a DIR* for a bftw_file.
+ * Open a bftw_file as a directory.
  *
  * @param cache
  *         The cache to hold the file.
@@ -467,39 +441,15 @@ static int bftw_file_open(struct bftw_cache *cache, struct bftw_file *file, cons
  * @param path
  *         The full path to the directory.
  * @return
- *         The opened DIR *, or NULL on error.
+ *         The opened directory, or NULL on error.
  */
-static DIR *bftw_file_opendir(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
+static struct bfs_dir *bftw_file_opendir(struct bftw_cache *cache, struct bftw_file *file, const char *path) {
 	int fd = bftw_file_open(cache, file, path);
 	if (fd < 0) {
 		return NULL;
 	}
 
-	// Now we dup() the fd and pass it to fdopendir().  This way we can
-	// close the DIR* as soon as we're done with it, reducing the memory
-	// footprint significantly, while keeping the fd around for future
-	// openat() calls.
-
-	int dfd = dup_cloexec(fd);
-
-	if (dfd < 0 && errno == EMFILE) {
-		if (bftw_cache_shrink(cache, file) == 0) {
-			dfd = dup_cloexec(fd);
-		}
-	}
-
-	if (dfd < 0) {
-		return NULL;
-	}
-
-	DIR *ret = fdopendir(dfd);
-	if (!ret) {
-		int error = errno;
-		close(dfd);
-		errno = error;
-	}
-
-	return ret;
+	return bfs_opendir(fd, NULL);
 }
 
 /** Free a bftw_file. */
@@ -610,70 +560,6 @@ static struct bftw_file **bftw_sort_files(struct bftw_file **head, struct bftw_f
 	return bftw_sort_merge(head, mid, tail);
 }
 
-/**
- * A directory reader.
- */
-struct bftw_reader {
-	/** The open handle to the directory. */
-	DIR *dir;
-	/** The current directory entry. */
-	struct dirent *de;
-	/** Any error code that has occurred. */
-	int error;
-};
-
-/** Initialize a reader. */
-static void bftw_reader_init(struct bftw_reader *reader) {
-	reader->dir = NULL;
-	reader->de = NULL;
-	reader->error = 0;
-}
-
-/** Open a directory for reading. */
-static int bftw_reader_open(struct bftw_reader *reader, struct bftw_cache *cache, struct bftw_file *file, const char *path) {
-	assert(!reader->dir);
-	assert(!reader->de);
-
-	reader->error = 0;
-
-	reader->dir = bftw_file_opendir(cache, file, path);
-	if (!reader->dir) {
-		reader->error = errno;
-		return -1;
-	}
-
-	return 0;
-}
-
-/** Read a directory entry. */
-static int bftw_reader_read(struct bftw_reader *reader) {
-	if (!reader->dir) {
-		return -1;
-	}
-
-	if (xreaddir(reader->dir, &reader->de) != 0) {
-		reader->error = errno;
-		return -1;
-	} else if (reader->de) {
-		return 1;
-	} else {
-		return 0;
-	}
-}
-
-/** Close a directory. */
-static int bftw_reader_close(struct bftw_reader *reader) {
-	int ret = 0;
-	if (reader->dir && closedir(reader->dir) != 0) {
-		reader->error = errno;
-		ret = -1;
-	}
-
-	reader->de = NULL;
-	reader->dir = NULL;
-	return ret;
-}
-
 /**
  * Holds the current state of the bftw() traversal.
  */
@@ -705,8 +591,15 @@ struct bftw_state {
 	struct bftw_file *file;
 	/** The previous file. */
 	struct bftw_file *previous;
-	/** The reader for the current directory. */
-	struct bftw_reader reader;
+
+	/** The currently open directory. */
+	struct bfs_dir *dir;
+	/** The current directory entry. */
+	struct bfs_dirent *de;
+	/** Storage for the directory entry. */
+	struct bfs_dirent de_storage;
+	/** Any error encountered while reading the directory. */
+	int direrror;
 
 	/** Extra data about the current file. */
 	struct BFTW ftwbuf;
@@ -729,7 +622,7 @@ static int bftw_state_init(struct bftw_state *state, const struct bftw_args *arg
 		goto err;
 	}
 
-	// Reserve 1 fd for the open DIR *
+	// Reserve 1 fd for the open bfs_dir
 	if (bftw_cache_init(&state->cache, args->nopenfd - 1) != 0) {
 		goto err;
 	}
@@ -744,7 +637,10 @@ static int bftw_state_init(struct bftw_state *state, const struct bftw_args *arg
 
 	state->file = NULL;
 	state->previous = NULL;
-	bftw_reader_init(&state->reader);
+
+	state->dir = NULL;
+	state->de = NULL;
+	state->direrror = 0;
 
 	return 0;
 
@@ -754,103 +650,6 @@ err:
 	return -1;
 }
 
-enum bftw_type bftw_mode_to_type(mode_t mode) {
-	switch (mode & S_IFMT) {
-#ifdef S_IFBLK
-	case S_IFBLK:
-		return BFTW_BLK;
-#endif
-#ifdef S_IFCHR
-	case S_IFCHR:
-		return BFTW_CHR;
-#endif
-#ifdef S_IFDIR
-	case S_IFDIR:
-		return BFTW_DIR;
-#endif
-#ifdef S_IFDOOR
-	case S_IFDOOR:
-		return BFTW_DOOR;
-#endif
-#ifdef S_IFIFO
-	case S_IFIFO:
-		return BFTW_FIFO;
-#endif
-#ifdef S_IFLNK
-	case S_IFLNK:
-		return BFTW_LNK;
-#endif
-#ifdef S_IFPORT
-	case S_IFPORT:
-		return BFTW_PORT;
-#endif
-#ifdef S_IFREG
-	case S_IFREG:
-		return BFTW_REG;
-#endif
-#ifdef S_IFSOCK
-	case S_IFSOCK:
-		return BFTW_SOCK;
-#endif
-#ifdef S_IFWHT
-	case S_IFWHT:
-		return BFTW_WHT;
-#endif
-
-	default:
-		return BFTW_UNKNOWN;
-	}
-}
-
-static enum bftw_type bftw_dirent_type(const struct dirent *de) {
-#if defined(_DIRENT_HAVE_D_TYPE) || defined(DT_UNKNOWN)
-	switch (de->d_type) {
-#ifdef DT_BLK
-	case DT_BLK:
-		return BFTW_BLK;
-#endif
-#ifdef DT_CHR
-	case DT_CHR:
-		return BFTW_CHR;
-#endif
-#ifdef DT_DIR
-	case DT_DIR:
-		return BFTW_DIR;
-#endif
-#ifdef DT_DOOR
-	case DT_DOOR:
-		return BFTW_DOOR;
-#endif
-#ifdef DT_FIFO
-	case DT_FIFO:
-		return BFTW_FIFO;
-#endif
-#ifdef DT_LNK
-	case DT_LNK:
-		return BFTW_LNK;
-#endif
-#ifdef DT_PORT
-	case DT_PORT:
-		return BFTW_PORT;
-#endif
-#ifdef DT_REG
-	case DT_REG:
-		return BFTW_REG;
-#endif
-#ifdef DT_SOCK
-	case DT_SOCK:
-		return BFTW_SOCK;
-#endif
-#ifdef DT_WHT
-	case DT_WHT:
-		return BFTW_WHT;
-#endif
-	}
-#endif
-
-	return BFTW_UNKNOWN;
-}
-
 /** Cached bfs_stat(). */
 static const struct bfs_stat *bftw_stat_impl(struct BFTW *ftwbuf, struct bftw_stat *cache, enum bfs_stat_flags flags) {
 	if (!cache->buf) {
@@ -898,20 +697,28 @@ const struct bfs_stat *bftw_cached_stat(const struct BFTW *ftwbuf, enum bfs_stat
 	}
 }
 
-enum bftw_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	if (ftwbuf->stat_flags & BFS_STAT_NOFOLLOW) {
-		if ((flags & BFS_STAT_NOFOLLOW) || ftwbuf->type != BFTW_LNK) {
+enum bfs_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
+	if (flags & BFS_STAT_NOFOLLOW) {
+		if (ftwbuf->type == BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
 			return ftwbuf->type;
 		}
-	} else if ((flags & (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW)) == BFS_STAT_TRYFOLLOW || ftwbuf->type == BFTW_LNK) {
-		return ftwbuf->type;
+	} else if (flags & BFS_STAT_TRYFOLLOW) {
+		if (ftwbuf->type != BFS_LNK || (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW)) {
+			return ftwbuf->type;
+		}
+	} else {
+		if (ftwbuf->type != BFS_LNK) {
+			return ftwbuf->type;
+		} else if (ftwbuf->stat_flags & BFS_STAT_TRYFOLLOW) {
+			return BFS_ERROR;
+		}
 	}
 
 	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, flags);
 	if (statbuf) {
-		return bftw_mode_to_type(statbuf->mode);
+		return bfs_mode_to_type(statbuf->mode);
 	} else {
-		return BFTW_ERROR;
+		return BFS_ERROR;
 	}
 }
 
@@ -946,15 +753,15 @@ static bool bftw_need_stat(const struct bftw_state *state) {
 	}
 
 	const struct BFTW *ftwbuf = &state->ftwbuf;
-	if (ftwbuf->type == BFTW_UNKNOWN) {
+	if (ftwbuf->type == BFS_UNKNOWN) {
 		return true;
 	}
 
-	if (ftwbuf->type == BFTW_LNK && !(ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
+	if (ftwbuf->type == BFS_LNK && !(ftwbuf->stat_flags & BFS_STAT_NOFOLLOW)) {
 		return true;
 	}
 
-	if (ftwbuf->type == BFTW_DIR) {
+	if (ftwbuf->type == BFS_DIR) {
 		if (state->flags & (BFTW_DETECT_CYCLES | BFTW_SKIP_MOUNTS | BFTW_PRUNE_MOUNTS)) {
 			return true;
 		}
@@ -1011,16 +818,15 @@ static int bftw_ensure_open(struct bftw_cache *cache, struct bftw_file *file, co
  */
 static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	struct bftw_file *file = state->file;
-	const struct bftw_reader *reader = &state->reader;
-	const struct dirent *de = reader->de;
+	const struct bfs_dirent *de = state->de;
 
 	struct BFTW *ftwbuf = &state->ftwbuf;
 	ftwbuf->path = state->path;
 	ftwbuf->root = file ? file->root->name : ftwbuf->path;
 	ftwbuf->depth = 0;
 	ftwbuf->visit = visit;
-	ftwbuf->type = BFTW_UNKNOWN;
-	ftwbuf->error = reader->error;
+	ftwbuf->type = BFS_UNKNOWN;
+	ftwbuf->error = state->direrror;
 	ftwbuf->at_fd = AT_FDCWD;
 	ftwbuf->at_path = ftwbuf->path;
 	ftwbuf->stat_flags = BFS_STAT_NOFOLLOW;
@@ -1031,7 +837,7 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	if (de) {
 		parent = file;
 		ftwbuf->depth = file->depth + 1;
-		ftwbuf->type = bftw_dirent_type(de);
+		ftwbuf->type = de->type;
 		ftwbuf->nameoff = bftw_child_nameoff(file);
 	} else if (file) {
 		parent = file->parent;
@@ -1056,7 +862,7 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	}
 
 	if (ftwbuf->error != 0) {
-		ftwbuf->type = BFTW_ERROR;
+		ftwbuf->type = BFS_ERROR;
 		return;
 	}
 
@@ -1073,18 +879,18 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	if (bftw_need_stat(state)) {
 		statbuf = bftw_stat(ftwbuf, ftwbuf->stat_flags);
 		if (statbuf) {
-			ftwbuf->type = bftw_mode_to_type(statbuf->mode);
+			ftwbuf->type = bfs_mode_to_type(statbuf->mode);
 		} else {
-			ftwbuf->type = BFTW_ERROR;
+			ftwbuf->type = BFS_ERROR;
 			ftwbuf->error = errno;
 			return;
 		}
 	}
 
-	if (ftwbuf->type == BFTW_DIR && (state->flags & BFTW_DETECT_CYCLES)) {
+	if (ftwbuf->type == BFS_DIR && (state->flags & BFTW_DETECT_CYCLES)) {
 		for (const struct bftw_file *ancestor = parent; ancestor; ancestor = ancestor->parent) {
 			if (ancestor->dev == statbuf->dev && ancestor->ino == statbuf->ino) {
-				ftwbuf->type = BFTW_ERROR;
+				ftwbuf->type = BFS_ERROR;
 				ftwbuf->error = ELOOP;
 				return;
 			}
@@ -1133,8 +939,8 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 	const struct BFTW *ftwbuf = &state->ftwbuf;
 	bftw_init_ftwbuf(state, visit);
 
-	// Never give the callback BFTW_ERROR unless BFTW_RECOVER is specified
-	if (ftwbuf->type == BFTW_ERROR && !(state->flags & BFTW_RECOVER)) {
+	// Never give the callback BFS_ERROR unless BFTW_RECOVER is specified
+	if (ftwbuf->type == BFS_ERROR && !(state->flags & BFTW_RECOVER)) {
 		state->error = ftwbuf->error;
 		return BFTW_STOP;
 	}
@@ -1155,7 +961,7 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 		return BFTW_STOP;
 	}
 
-	if (visit != BFTW_PRE || ftwbuf->type != BFTW_DIR) {
+	if (visit != BFTW_PRE || ftwbuf->type != BFS_DIR) {
 		ret = BFTW_PRUNE;
 		goto done;
 	}
@@ -1184,9 +990,8 @@ static int bftw_push(struct bftw_state *state, const char *name, bool fill_id) {
 		return -1;
 	}
 
-	struct dirent *de = state->reader.de;
-	if (de) {
-		file->type = bftw_dirent_type(de);
+	if (state->de) {
+		file->type = state->de->type;
 	}
 
 	if (fill_id) {
@@ -1251,18 +1056,45 @@ static int bftw_pop(struct bftw_state *state) {
 }
 
 /**
- * Open a reader for the current directory.
+ * Open the current directory.
  */
-static struct bftw_reader *bftw_open(struct bftw_state *state) {
-	struct bftw_reader *reader = &state->reader;
-	bftw_reader_open(reader, &state->cache, state->file, state->path);
-	return reader;
+static void bftw_opendir(struct bftw_state *state) {
+	assert(!state->dir);
+	assert(!state->de);
+
+	state->direrror = 0;
+
+	state->dir = bftw_file_opendir(&state->cache, state->file, state->path);
+	if (!state->dir) {
+		state->direrror = errno;
+	}
 }
 
 /**
- * Flags controlling which files get visited when releasing a reader/file.
+ * Read an entry from the current directory.
  */
-enum bftw_release_flags {
+static int bftw_readdir(struct bftw_state *state) {
+	if (!state->dir) {
+		return -1;
+	}
+
+	int ret = bfs_readdir(state->dir, &state->de_storage);
+	if (ret > 0) {
+		state->de = &state->de_storage;
+	} else if (ret == 0) {
+		state->de = NULL;
+	} else {
+		state->de = NULL;
+		state->direrror = errno;
+	}
+
+	return ret;
+}
+
+/**
+ * Flags controlling which files get visited when done with a directory.
+ */
+enum bftw_gc_flags {
 	/** Don't visit anything. */
 	BFTW_VISIT_NONE = 0,
 	/** Visit the file itself. */
@@ -1274,21 +1106,38 @@ enum bftw_release_flags {
 };
 
 /**
- * Close and release the reader.
+ * Close the current directory.
  */
-static enum bftw_action bftw_release_reader(struct bftw_state *state, enum bftw_release_flags flags) {
+static enum bftw_action bftw_closedir(struct bftw_state *state, enum bftw_gc_flags flags) {
+	struct bftw_file *file = state->file;
 	enum bftw_action ret = BFTW_CONTINUE;
 
-	struct bftw_reader *reader = &state->reader;
-	bftw_reader_close(reader);
+	if (state->dir) {
+		assert(file->fd >= 0);
 
-	if (reader->error != 0) {
+		if (file->refcount > 1) {
+			// Keep the fd around if any subdirectories exist
+			file->fd = bfs_freedir(state->dir);
+		} else {
+			bfs_closedir(state->dir);
+			file->fd = -1;
+		}
+
+		if (file->fd < 0) {
+			bftw_cache_remove(&state->cache, file);
+		}
+	}
+
+	state->de = NULL;
+	state->dir = NULL;
+
+	if (state->direrror != 0) {
 		if (flags & BFTW_VISIT_FILE) {
 			ret = bftw_visit(state, NULL, BFTW_PRE);
 		} else {
-			state->error = reader->error;
+			state->error = state->direrror;
 		}
-		reader->error = 0;
+		state->direrror = 0;
 	}
 
 	return ret;
@@ -1297,7 +1146,7 @@ static enum bftw_action bftw_release_reader(struct bftw_state *state, enum bftw_
 /**
  * Finalize and free a file we're done with.
  */
-static enum bftw_action bftw_release_file(struct bftw_state *state, enum bftw_release_flags flags) {
+static enum bftw_action bftw_gc_file(struct bftw_state *state, enum bftw_gc_flags flags) {
 	enum bftw_action ret = BFTW_CONTINUE;
 
 	if (!(state->flags & BFTW_POST_ORDER)) {
@@ -1334,7 +1183,7 @@ static enum bftw_action bftw_release_file(struct bftw_state *state, enum bftw_re
 static void bftw_drain_queue(struct bftw_state *state, struct bftw_queue *queue) {
 	while (queue->head) {
 		state->file = bftw_queue_pop(queue);
-		bftw_release_file(state, BFTW_VISIT_NONE);
+		bftw_gc_file(state, BFTW_VISIT_NONE);
 	}
 }
 
@@ -1347,9 +1196,9 @@ static void bftw_drain_queue(struct bftw_state *state, struct bftw_queue *queue)
 static int bftw_state_destroy(struct bftw_state *state) {
 	dstrfree(state->path);
 
-	bftw_release_reader(state, BFTW_VISIT_NONE);
+	bftw_closedir(state, BFTW_VISIT_NONE);
 
-	bftw_release_file(state, BFTW_VISIT_NONE);
+	bftw_gc_file(state, BFTW_VISIT_NONE);
 	bftw_drain_queue(state, &state->queue);
 
 	bftw_cache_destroy(&state->cache);
@@ -1404,11 +1253,11 @@ static int bftw_stream(const struct bftw_args *args) {
 	bftw_batch_finish(&state);
 
 	while (bftw_pop(&state) > 0) {
-		struct bftw_reader *reader = bftw_open(&state);
+		bftw_opendir(&state);
 
 		bftw_batch_start(&state);
-		while (bftw_reader_read(reader) > 0) {
-			const char *name = reader->de->d_name;
+		while (bftw_readdir(&state) > 0) {
+			const char *name = state.de->name;
 
 			switch (bftw_visit(&state, name, BFTW_PRE)) {
 			case BFTW_CONTINUE:
@@ -1425,10 +1274,10 @@ static int bftw_stream(const struct bftw_args *args) {
 		}
 		bftw_batch_finish(&state);
 
-		if (bftw_release_reader(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
+		if (bftw_closedir(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
 			goto done;
 		}
-		if (bftw_release_file(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
+		if (bftw_gc_file(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
 			goto done;
 		}
 	}
@@ -1455,34 +1304,34 @@ static int bftw_batch(const struct bftw_args *args) {
 	bftw_batch_finish(&state);
 
 	while (bftw_pop(&state) > 0) {
-		enum bftw_release_flags relflags = BFTW_VISIT_ALL;
+		enum bftw_gc_flags gcflags = BFTW_VISIT_ALL;
 
 		switch (bftw_visit(&state, NULL, BFTW_PRE)) {
 		case BFTW_CONTINUE:
 			break;
 		case BFTW_PRUNE:
-			relflags &= ~BFTW_VISIT_FILE;
+			gcflags &= ~BFTW_VISIT_FILE;
 			goto next;
 		case BFTW_STOP:
 			goto done;
 		}
 
-		struct bftw_reader *reader = bftw_open(&state);
+		bftw_opendir(&state);
 
 		bftw_batch_start(&state);
-		while (bftw_reader_read(reader) > 0) {
-			if (bftw_push(&state, reader->de->d_name, false) != 0) {
+		while (bftw_readdir(&state) > 0) {
+			if (bftw_push(&state, state.de->name, false) != 0) {
 				goto done;
 			}
 		}
 		bftw_batch_finish(&state);
 
-		if (bftw_release_reader(&state, relflags) == BFTW_STOP) {
+		if (bftw_closedir(&state, gcflags) == BFTW_STOP) {
 			goto done;
 		}
 
 	next:
-		if (bftw_release_file(&state, relflags) == BFTW_STOP) {
+		if (bftw_gc_file(&state, gcflags) == BFTW_STOP) {
 			goto done;
 		}
 	}
@@ -1535,7 +1384,7 @@ static enum bftw_action bftw_ids_callback(const struct BFTW *ftwbuf, void *ptr)
 		mutbuf->visit = state->visit;
 	}
 
-	if (ftwbuf->type == BFTW_ERROR) {
+	if (ftwbuf->type == BFS_ERROR) {
 		if (ftwbuf->depth + 1 >= state->min_depth) {
 			return state->delegate(ftwbuf, state->ptr);
 		} else {
@@ -1562,13 +1411,13 @@ static enum bftw_action bftw_ids_callback(const struct BFTW *ftwbuf, void *ptr)
 
 	switch (ret) {
 	case BFTW_CONTINUE:
-		if (ftwbuf->type == BFTW_DIR && ftwbuf->depth + 1 >= state->max_depth) {
+		if (ftwbuf->type == BFS_DIR && ftwbuf->depth + 1 >= state->max_depth) {
 			state->bottom = false;
 			ret = BFTW_PRUNE;
 		}
 		break;
 	case BFTW_PRUNE:
-		if (ftwbuf->type == BFTW_DIR) {
+		if (ftwbuf->type == BFS_DIR) {
 			if (!trie_insert_str(&state->pruned, ftwbuf->path)) {
 				state->error = errno;
 				state->quit = true;
