@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -21,45 +21,9 @@
 #ifndef BFS_BFTW_H
 #define BFS_BFTW_H
 
-#include "mtab.h"
+#include "dir.h"
 #include "stat.h"
 #include <stddef.h>
-#include <sys/types.h>
-
-/**
- * Possible file types.
- */
-enum bftw_typeflag {
-	/** Unknown type. */
-	BFTW_UNKNOWN = 0,
-	/** Block device. */
-	BFTW_BLK     = 1 << 0,
-	/** Character device. */
-	BFTW_CHR     = 1 << 1,
-	/** Directory. */
-	BFTW_DIR     = 1 << 2,
-	/** Solaris door. */
-	BFTW_DOOR    = 1 << 3,
-	/** Pipe. */
-	BFTW_FIFO    = 1 << 4,
-	/** Symbolic link. */
-	BFTW_LNK     = 1 << 5,
-	/** Solaris event port. */
-	BFTW_PORT    = 1 << 6,
-	/** Regular file. */
-	BFTW_REG     = 1 << 7,
-	/** Socket. */
-	BFTW_SOCK    = 1 << 8,
-	/** BSD whiteout. */
-	BFTW_WHT     = 1 << 9,
-	/** An error occurred for this file. */
-	BFTW_ERROR   = 1 << 10,
-};
-
-/**
- * Convert a bfs_stat() mode to a bftw() typeflag.
- */
-enum bftw_typeflag bftw_mode_typeflag(mode_t mode);
 
 /**
  * Possible visit occurrences.
@@ -100,8 +64,8 @@ struct BFTW {
 	enum bftw_visit visit;
 
 	/** The file type. */
-	enum bftw_typeflag typeflag;
-	/** The errno that occurred, if typeflag == BFTW_ERROR. */
+	enum bfs_type type;
+	/** The errno that occurred, if type == BFTW_ERROR. */
 	int error;
 
 	/** A parent file descriptor for the *at() family of calls. */
@@ -110,7 +74,7 @@ struct BFTW {
 	const char *at_path;
 
 	/** Flags for bfs_stat(). */
-	enum bfs_stat_flag stat_flags;
+	enum bfs_stat_flags stat_flags;
 	/** Cached bfs_stat() info for BFS_STAT_NOFOLLOW. */
 	struct bftw_stat lstat_cache;
 	/** Cached bfs_stat() info for BFS_STAT_FOLLOW. */
@@ -128,7 +92,20 @@ struct BFTW {
  * @return
  *         A pointer to a bfs_stat() buffer, or NULL if the call failed.
  */
-const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flag flags);
+const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags);
+
+/**
+ * Get bfs_stat() info for a file encountered during bftw(), if it has already
+ * been cached.
+ *
+ * @param ftwbuf
+ *         bftw() data for the file to stat.
+ * @param flags
+ *         flags for bfs_stat().  Pass ftwbuf->stat_flags for the default flags.
+ * @return
+ *         A pointer to a bfs_stat() buffer, or NULL if no stat info is cached.
+ */
+const struct bfs_stat *bftw_cached_stat(const struct BFTW *ftwbuf, enum bfs_stat_flags flags);
 
 /**
  * Get the type of a file encountered during bftw(), with flags controlling
@@ -142,7 +119,7 @@ const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flag f
  * @return
  *         The type of the file, or BFTW_ERROR if an error occurred.
  */
-enum bftw_typeflag bftw_typeflag(const struct BFTW *ftwbuf, enum bfs_stat_flag flags);
+enum bfs_type bftw_type(const struct BFTW *ftwbuf, enum bfs_stat_flags flags);
 
 /**
  * Walk actions returned by the bftw() callback.
@@ -177,17 +154,17 @@ enum bftw_flags {
 	/** Attempt to recover from encountered errors. */
 	BFTW_RECOVER       = 1 << 1,
 	/** Visit directories in post-order as well as pre-order. */
-	BFTW_DEPTH         = 1 << 2,
+	BFTW_POST_ORDER    = 1 << 2,
 	/** If the initial path is a symbolic link, follow it. */
-	BFTW_COMFOLLOW     = 1 << 3,
+	BFTW_FOLLOW_ROOTS  = 1 << 3,
 	/** Follow all symbolic links. */
-	BFTW_LOGICAL       = 1 << 4,
+	BFTW_FOLLOW_ALL    = 1 << 4,
 	/** Detect directory cycles. */
 	BFTW_DETECT_CYCLES = 1 << 5,
 	/** Skip mount points and their descendents. */
-	BFTW_MOUNT         = 1 << 6,
+	BFTW_SKIP_MOUNTS   = 1 << 6,
 	/** Skip the descendents of mount points. */
-	BFTW_XDEV          = 1 << 7,
+	BFTW_PRUNE_MOUNTS  = 1 << 7,
 	/** Sort directory entries before processing them. */
 	BFTW_SORT          = 1 << 8,
 };
@@ -202,6 +179,8 @@ enum bftw_strategy {
 	BFTW_DFS,
 	/** Iterative deepening search. */
 	BFTW_IDS,
+	/** Exponential deepening search. */
+	BFTW_EDS,
 };
 
 /**
