@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -16,6 +16,7 @@
 
 #include "color.h"
 #include "bftw.h"
+#include "dir.h"
 #include "dstring.h"
 #include "expr.h"
 #include "fsade.h"
@@ -24,12 +25,13 @@
 #include "util.h"
 #include <assert.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <stdarg.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/stat.h>
+#include <time.h>
 #include <unistd.h>
 
 /**
@@ -494,13 +496,31 @@ void free_colors(struct colors *colors) {
 	}
 }
 
-CFILE *cfopen(const char *path, const struct colors *colors) {
+static CFILE *cfalloc(void) {
 	CFILE *cfile = malloc(sizeof(*cfile));
 	if (!cfile) {
 		return NULL;
 	}
 
+	cfile->buffer = dstralloc(128);
+	if (!cfile->buffer) {
+		free(cfile);
+		return NULL;
+	}
+
+	cfile->file = NULL;
+	cfile->colors = NULL;
 	cfile->close = false;
+
+	return cfile;
+}
+
+CFILE *cfopen(const char *path, const struct colors *colors) {
+	CFILE *cfile = cfalloc();
+	if (!cfile) {
+		return NULL;
+	}
+
 	cfile->file = fopen(path, "wb");
 	if (!cfile->file) {
 		cfclose(cfile);
@@ -518,7 +538,7 @@ CFILE *cfopen(const char *path, const struct colors *colors) {
 }
 
 CFILE *cfdup(FILE *file, const struct colors *colors) {
-	CFILE *cfile = malloc(sizeof(*cfile));
+	CFILE *cfile = cfalloc();
 	if (!cfile) {
 		return NULL;
 	}
@@ -537,12 +557,17 @@ CFILE *cfdup(FILE *file, const struct colors *colors) {
 
 int cfclose(CFILE *cfile) {
 	int ret = 0;
+
 	if (cfile) {
+		dstrfree(cfile->buffer);
+
 		if (cfile->close) {
 			ret = fclose(cfile->file);
 		}
+
 		free(cfile);
 	}
+
 	return ret;
 }
 
@@ -557,8 +582,8 @@ static bool is_link_broken(const struct BFTW *ftwbuf) {
 
 /** Get the color for a file. */
 static const char *file_color(const struct colors *colors, const char *filename, const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	enum bftw_type type = bftw_type(ftwbuf, flags);
-	if (type == BFTW_ERROR) {
+	enum bfs_type type = bftw_type(ftwbuf, flags);
+	if (type == BFS_ERROR) {
 		goto error;
 	}
 
@@ -566,7 +591,7 @@ static const char *file_color(const struct colors *colors, const char *filename,
 	const char *color = NULL;
 
 	switch (type) {
-	case BFTW_REG:
+	case BFS_REG:
 		if (colors->setuid || colors->setgid || colors->executable || colors->multi_hard) {
 			statbuf = bftw_stat(ftwbuf, flags);
 			if (!statbuf) {
@@ -596,7 +621,7 @@ static const char *file_color(const struct colors *colors, const char *filename,
 
 		break;
 
-	case BFTW_DIR:
+	case BFS_DIR:
 		if (colors->sticky_other_writable || colors->other_writable || colors->sticky) {
 			statbuf = bftw_stat(ftwbuf, flags);
 			if (!statbuf) {
@@ -616,7 +641,7 @@ static const char *file_color(const struct colors *colors, const char *filename,
 
 		break;
 
-	case BFTW_LNK:
+	case BFS_LNK:
 		if (colors->orphan && is_link_broken(ftwbuf)) {
 			color = colors->orphan;
 		} else {
@@ -624,19 +649,19 @@ static const char *file_color(const struct colors *colors, const char *filename,
 		}
 		break;
 
-	case BFTW_BLK:
+	case BFS_BLK:
 		color = colors->blockdev;
 		break;
-	case BFTW_CHR:
+	case BFS_CHR:
 		color = colors->chardev;
 		break;
-	case BFTW_FIFO:
+	case BFS_FIFO:
 		color = colors->pipe;
 		break;
-	case BFTW_SOCK:
+	case BFS_SOCK:
 		color = colors->socket;
 		break;
-	case BFTW_DOOR:
+	case BFS_DOOR:
 		color = colors->door;
 		break;
 
@@ -658,29 +683,17 @@ error:
 	}
 }
 
-/** Print a fixed-length string. */
-static int print_strn(const char *str, size_t len, FILE *file) {
-	if (fwrite(str, 1, len, file) == len) {
-		return 0;
-	} else {
-		return -1;
-	}
-}
-
-/** Print a dstring. */
-static int print_dstr(const char *str, FILE *file) {
-	return print_strn(str, dstrlen(str), file);
-}
-
 /** Print an ANSI escape sequence. */
-static int print_esc(const struct colors *colors, const char *esc, FILE *file) {
-	if (print_dstr(colors->leftcode, file) != 0) {
+static int print_esc(CFILE *cfile, const char *esc) {
+	const struct colors *colors = cfile->colors;
+
+	if (dstrdcat(&cfile->buffer, colors->leftcode) != 0) {
 		return -1;
 	}
-	if (print_dstr(esc, file) != 0) {
+	if (dstrdcat(&cfile->buffer, esc) != 0) {
 		return -1;
 	}
-	if (print_dstr(colors->rightcode, file) != 0) {
+	if (dstrdcat(&cfile->buffer, colors->rightcode) != 0) {
 		return -1;
 	}
 
@@ -688,26 +701,28 @@ static int print_esc(const struct colors *colors, const char *esc, FILE *file) {
 }
 
 /** Reset after an ANSI escape sequence. */
-static int print_reset(const struct colors *colors, FILE *file) {
+static int print_reset(CFILE *cfile) {
+	const struct colors *colors = cfile->colors;
+
 	if (colors->endcode) {
-		return print_dstr(colors->endcode, file);
+		return dstrdcat(&cfile->buffer, colors->endcode);
 	} else {
-		return print_esc(colors, colors->reset, file);
+		return print_esc(cfile, colors->reset);
 	}
 }
 
 /** Print a string with an optional color. */
-static int print_colored(const struct colors *colors, const char *esc, const char *str, size_t len, FILE *file) {
+static int print_colored(CFILE *cfile, const char *esc, const char *str, size_t len) {
 	if (esc) {
-		if (print_esc(colors, esc, file) != 0) {
+		if (print_esc(cfile, esc) != 0) {
 			return -1;
 		}
 	}
-	if (print_strn(str, len, file) != 0) {
+	if (dstrncat(&cfile->buffer, str, len) != 0) {
 		return -1;
 	}
 	if (esc) {
-		if (print_reset(colors, file) != 0) {
+		if (print_reset(cfile) != 0) {
 			return -1;
 		}
 	}
@@ -719,7 +734,7 @@ static int print_colored(const struct colors *colors, const char *esc, const cha
 static ssize_t first_broken_offset(const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags, size_t max) {
 	ssize_t ret = max;
 
-	if (bftw_type(ftwbuf, flags) != BFTW_ERROR) {
+	if (bftw_type(ftwbuf, flags) != BFS_ERROR) {
 		goto out;
 	}
 
@@ -777,7 +792,6 @@ out:
 /** Print the directories leading up to a file. */
 static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags, size_t nameoff) {
 	const struct colors *colors = cfile->colors;
-	FILE *file = cfile->file;
 
 	ssize_t broken = first_broken_offset(path, ftwbuf, flags, nameoff);
 	if (broken < 0) {
@@ -785,7 +799,7 @@ static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW
 	}
 
 	if (broken > 0) {
-		if (print_colored(colors, colors->directory, path, broken, file) != 0) {
+		if (print_colored(cfile, colors->directory, path, broken) != 0) {
 			return -1;
 		}
 	}
@@ -795,7 +809,7 @@ static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW
 		if (!color) {
 			color = colors->orphan;
 		}
-		if (print_colored(colors, color, path + broken, nameoff - broken, file) != 0) {
+		if (print_colored(cfile, color, path + broken, nameoff - broken) != 0) {
 			return -1;
 		}
 	}
@@ -805,9 +819,6 @@ static int print_dirs_colored(CFILE *cfile, const char *path, const struct BFTW
 
 /** Print a path with colors. */
 static int print_path_colored(CFILE *cfile, const char *path, const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
-	const struct colors *colors = cfile->colors;
-	FILE *file = cfile->file;
-
 	size_t nameoff;
 	if (path == ftwbuf->path) {
 		nameoff = ftwbuf->nameoff;
@@ -818,19 +829,19 @@ static int print_path_colored(CFILE *cfile, const char *path, const struct BFTW
 	print_dirs_colored(cfile, path, ftwbuf, flags, nameoff);
 
 	const char *filename = path + nameoff;
-	const char *color = file_color(colors, filename, ftwbuf, flags);
-	return print_colored(colors, color, filename, strlen(filename), file);
+	const char *color = file_color(cfile->colors, filename, ftwbuf, flags);
+	return print_colored(cfile, color, filename, strlen(filename));
 }
 
 /** Print the path to a file with the appropriate colors. */
 static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 	const struct colors *colors = cfile->colors;
 	if (!colors) {
-		return fputs(ftwbuf->path, cfile->file) == EOF ? -1 : 0;
+		return dstrcat(&cfile->buffer, ftwbuf->path);
 	}
 
 	enum bfs_stat_flags flags = ftwbuf->stat_flags;
-	if (colors && colors->link_as_target && ftwbuf->type == BFTW_LNK) {
+	if (colors && colors->link_as_target && ftwbuf->type == BFS_LNK) {
 		flags = BFS_STAT_TRYFOLLOW;
 	}
 
@@ -839,46 +850,47 @@ static int print_path(CFILE *cfile, const struct BFTW *ftwbuf) {
 
 /** Print a link target with the appropriate colors. */
 static int print_link_target(CFILE *cfile, const struct BFTW *ftwbuf) {
-	int ret = -1;
-
 	const struct bfs_stat *statbuf = bftw_cached_stat(ftwbuf, BFS_STAT_NOFOLLOW);
 	size_t len = statbuf ? statbuf->size : 0;
 
 	char *target = xreadlinkat(ftwbuf->at_fd, ftwbuf->at_path, len);
 	if (!target) {
-		goto done;
+		return -1;
 	}
 
-	if (!cfile->colors) {
-		ret = fputs(target, cfile->file) == EOF ? -1 : 0;
-		goto done;
+	int ret;
+	if (cfile->colors) {
+		ret = print_path_colored(cfile, target, ftwbuf, BFS_STAT_FOLLOW);
+	} else {
+		ret = dstrcat(&cfile->buffer, target);
 	}
 
-	ret = print_path_colored(cfile, target, ftwbuf, BFS_STAT_FOLLOW);
-
-done:
 	free(target);
 	return ret;
 }
 
+/** Format some colored output to the buffer. */
+BFS_FORMATTER(2, 3)
+static int cbuff(CFILE *cfile, const char *format, ...);
+
 /** Dump a parsed expression tree, for debugging. */
 static int print_expr(CFILE *cfile, const struct expr *expr, bool verbose) {
-	if (fputs("(", cfile->file) == EOF) {
+	if (dstrcat(&cfile->buffer, "(") != 0) {
 		return -1;
 	}
 
 	if (expr->lhs || expr->rhs) {
-		if (cfprintf(cfile, "${red}%s${rs}", expr->argv[0]) < 0) {
+		if (cbuff(cfile, "${red}%s${rs}", expr->argv[0]) < 0) {
 			return -1;
 		}
 	} else {
-		if (cfprintf(cfile, "${blu}%s${rs}", expr->argv[0]) < 0) {
+		if (cbuff(cfile, "${blu}%s${rs}", expr->argv[0]) < 0) {
 			return -1;
 		}
 	}
 
 	for (size_t i = 1; i < expr->argc; ++i) {
-		if (cfprintf(cfile, " ${bld}%s${rs}", expr->argv[i]) < 0) {
+		if (cbuff(cfile, " ${bld}%s${rs}", expr->argv[i]) < 0) {
 			return -1;
 		}
 	}
@@ -889,14 +901,14 @@ static int print_expr(CFILE *cfile, const struct expr *expr, bool verbose) {
 			rate = 100.0*expr->successes/expr->evaluations;
 			time = (1.0e9*expr->elapsed.tv_sec + expr->elapsed.tv_nsec)/expr->evaluations;
 		}
-		if (cfprintf(cfile, " [${ylw}%zu${rs}/${ylw}%zu${rs}=${ylw}%g%%${rs}; ${ylw}%gns${rs}]",
-		             expr->successes, expr->evaluations, rate, time)) {
+		if (cbuff(cfile, " [${ylw}%zu${rs}/${ylw}%zu${rs}=${ylw}%g%%${rs}; ${ylw}%gns${rs}]",
+		          expr->successes, expr->evaluations, rate, time)) {
 			return -1;
 		}
 	}
 
 	if (expr->lhs) {
-		if (fputs(" ", cfile->file) == EOF) {
+		if (dstrcat(&cfile->buffer, " ") != 0) {
 			return -1;
 		}
 		if (print_expr(cfile, expr->lhs, verbose) != 0) {
@@ -905,7 +917,7 @@ static int print_expr(CFILE *cfile, const struct expr *expr, bool verbose) {
 	}
 
 	if (expr->rhs) {
-		if (fputs(" ", cfile->file) == EOF) {
+		if (dstrcat(&cfile->buffer, " ") != 0) {
 			return -1;
 		}
 		if (print_expr(cfile, expr->rhs, verbose) != 0) {
@@ -913,29 +925,20 @@ static int print_expr(CFILE *cfile, const struct expr *expr, bool verbose) {
 		}
 	}
 
-	if (fputs(")", cfile->file) == EOF) {
+	if (dstrcat(&cfile->buffer, ")") != 0) {
 		return -1;
 	}
 
 	return 0;
 }
 
-int cfprintf(CFILE *cfile, const char *format, ...) {
-	va_list args;
-	va_start(args, format);
-	int ret = cvfprintf(cfile, format, args);
-	va_end(args);
-	return ret;
-}
-
-int cvfprintf(CFILE *cfile, const char *format, va_list args) {
+static int cvbuff(CFILE *cfile, const char *format, va_list args) {
 	const struct colors *colors = cfile->colors;
-	FILE *file = cfile->file;
 	int error = errno;
 
 	for (const char *i = format; *i; ++i) {
 		size_t verbatim = strcspn(i, "%$");
-		if (fwrite(i, 1, verbatim, file) != verbatim) {
+		if (dstrncat(&cfile->buffer, i, verbatim) != 0) {
 			return -1;
 		}
 
@@ -944,31 +947,31 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 		case '%':
 			switch (*++i) {
 			case '%':
-				if (fputc('%', file) == EOF) {
+				if (dstrapp(&cfile->buffer, '%') != 0) {
 					return -1;
 				}
 				break;
 
 			case 'c':
-				if (fputc(va_arg(args, int), file) == EOF) {
+				if (dstrapp(&cfile->buffer, va_arg(args, int)) != 0) {
 					return -1;
 				}
 				break;
 
 			case 'd':
-				if (fprintf(file, "%d", va_arg(args, int)) < 0) {
+				if (dstrcatf(&cfile->buffer, "%d", va_arg(args, int)) != 0) {
 					return -1;
 				}
 				break;
 
 			case 'g':
-				if (fprintf(file, "%g", va_arg(args, double)) < 0) {
+				if (dstrcatf(&cfile->buffer, "%g", va_arg(args, double)) != 0) {
 					return -1;
 				}
 				break;
 
 			case 's':
-				if (fputs(va_arg(args, const char *), file) == EOF) {
+				if (dstrcat(&cfile->buffer, va_arg(args, const char *)) != 0) {
 					return -1;
 				}
 				break;
@@ -978,13 +981,13 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 				if (*i != 'u') {
 					goto invalid;
 				}
-				if (fprintf(file, "%zu", va_arg(args, size_t)) < 0) {
+				if (dstrcatf(&cfile->buffer, "%zu", va_arg(args, size_t)) != 0) {
 					return -1;
 				}
 				break;
 
 			case 'm':
-				if (fputs(strerror(error), file) == EOF) {
+				if (dstrcat(&cfile->buffer, strerror(error)) != 0) {
 					return -1;
 				}
 				break;
@@ -1028,7 +1031,7 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 		case '$':
 			switch (*++i) {
 			case '$':
-				if (fputc('$', file) == EOF) {
+				if (dstrapp(&cfile->buffer, '$') != 0) {
 					return -1;
 				}
 				break;
@@ -1054,7 +1057,7 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 					goto invalid;
 				}
 				if (*esc) {
-					if (print_esc(colors, *esc, file) != 0) {
+					if (print_esc(cfile, *esc) != 0) {
 						return -1;
 					}
 				}
@@ -1071,13 +1074,43 @@ int cvfprintf(CFILE *cfile, const char *format, va_list args) {
 		default:
 			return 0;
 		}
-
 	}
 
 	return 0;
 
 invalid:
-	assert(false);
+	assert(!"Invalid format string");
 	errno = EINVAL;
 	return -1;
 }
+
+static int cbuff(CFILE *cfile, const char *format, ...) {
+	va_list args;
+	va_start(args, format);
+	int ret = cvbuff(cfile, format, args);
+	va_end(args);
+	return ret;
+}
+
+int cvfprintf(CFILE *cfile, const char *format, va_list args) {
+	assert(dstrlen(cfile->buffer) == 0);
+
+	int ret = -1;
+	if (cvbuff(cfile, format, args) == 0) {
+		size_t len = dstrlen(cfile->buffer);
+		if (fwrite(cfile->buffer, 1, len, cfile->file) == len) {
+			ret = 0;
+		}
+	}
+
+	dstresize(&cfile->buffer, 0);
+	return ret;
+}
+
+int cfprintf(CFILE *cfile, const char *format, ...) {
+	va_list args;
+	va_start(args, format);
+	int ret = cvfprintf(cfile, format, args);
+	va_end(args);
+	return ret;
+}
