@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2020 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -22,8 +22,10 @@
 #include "bar.h"
 #include "bftw.h"
 #include "color.h"
+#include "ctx.h"
 #include "darray.h"
 #include "diag.h"
+#include "dir.h"
 #include "dstring.h"
 #include "exec.h"
 #include "expr.h"
@@ -36,12 +38,12 @@
 #include "trie.h"
 #include "util.h"
 #include <assert.h>
-#include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <fnmatch.h>
 #include <grp.h>
 #include <pwd.h>
+#include <regex.h>
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -71,6 +73,9 @@ struct eval_state {
  */
 BFS_FORMATTER(2, 3)
 static void eval_error(struct eval_state *state, const char *format, ...) {
+	// By POSIX, any errors should be accompanied by a non-zero exit status
+	*state->ret = EXIT_FAILURE;
+
 	int error = errno;
 	const struct bfs_ctx *ctx = state->ctx;
 	CFILE *cerr = ctx->cerr;
@@ -99,7 +104,6 @@ static bool eval_should_ignore(const struct eval_state *state, int error) {
 static void eval_report_error(struct eval_state *state) {
 	if (!eval_should_ignore(state, errno)) {
 		eval_error(state, "%m.\n");
-		*state->ret = EXIT_FAILURE;
 	}
 }
 
@@ -194,7 +198,6 @@ static const struct timespec *eval_stat_time(const struct bfs_stat *statbuf, enu
 	const struct timespec *ret = bfs_stat_time(statbuf, field);
 	if (!ret) {
 		eval_error(state, "Couldn't get file %s: %m.\n", bfs_stat_field_name(field));
-		*state->ret = EXIT_FAILURE;
 	}
 	return ret;
 }
@@ -233,11 +236,13 @@ bool eval_time(const struct expr *expr, struct eval_state *state) {
 
 	time_t diff = timespec_diff(&expr->reftime, time);
 	switch (expr->time_unit) {
+	case DAYS:
+		diff /= 60*24;
+		fallthrough;
 	case MINUTES:
 		diff /= 60;
-		break;
-	case DAYS:
-		diff /= 60*60*24;
+		fallthrough;
+	case SECONDS:
 		break;
 	}
 
@@ -259,8 +264,13 @@ bool eval_used(const struct expr *expr, struct eval_state *state) {
 		return false;
 	}
 
-	time_t diff = timespec_diff(atime, ctime);
-	diff /= 60*60*24;
+	long long diff = timespec_diff(atime, ctime);
+	if (diff < 0) {
+		return false;
+	}
+
+	long long day_seconds = 60*60*24;
+	diff = (diff + day_seconds - 1) / day_seconds;
 	return expr_cmp(expr, diff);
 }
 
@@ -338,10 +348,10 @@ bool eval_delete(const struct expr *expr, struct eval_state *state) {
 	int flag = 0;
 
 	// We need to know the actual type of the path, not what it points to
-	enum bftw_type type = bftw_type(ftwbuf, BFS_STAT_NOFOLLOW);
-	if (type == BFTW_DIR) {
+	enum bfs_type type = bftw_type(ftwbuf, BFS_STAT_NOFOLLOW);
+	if (type == BFS_DIR) {
 		flag |= AT_REMOVEDIR;
-	} else if (type == BFTW_ERROR) {
+	} else if (type == BFS_ERROR) {
 		eval_report_error(state);
 		return false;
 	}
@@ -379,7 +389,6 @@ bool eval_exec(const struct expr *expr, struct eval_state *state) {
 	bool ret = bfs_exec(expr->execbuf, state->ftwbuf) == 0;
 	if (errno != 0) {
 		eval_error(state, "%s %s: %m.\n", expr->argv[0], expr->argv[1]);
-		*state->ret = EXIT_FAILURE;
 	}
 	return ret;
 }
@@ -408,29 +417,22 @@ bool eval_empty(const struct expr *expr, struct eval_state *state) {
 	bool ret = false;
 	const struct BFTW *ftwbuf = state->ftwbuf;
 
-	if (ftwbuf->type == BFTW_DIR) {
-		int dfd = openat(ftwbuf->at_fd, ftwbuf->at_path, O_RDONLY | O_CLOEXEC | O_DIRECTORY);
-		if (dfd < 0) {
-			eval_report_error(state);
-			goto done;
-		}
-
-		DIR *dir = fdopendir(dfd);
+	if (ftwbuf->type == BFS_DIR) {
+		struct bfs_dir *dir = bfs_opendir(ftwbuf->at_fd, ftwbuf->at_path);
 		if (!dir) {
 			eval_report_error(state);
-			close(dfd);
 			goto done;
 		}
 
-		struct dirent *de;
-		if (xreaddir(dir, &de) == 0) {
-			ret = !de;
-		} else {
+		int did_read = bfs_readdir(dir, NULL);
+		if (did_read < 0) {
 			eval_report_error(state);
+		} else {
+			ret = !did_read;
 		}
 
-		closedir(dir);
-	} else if (ftwbuf->type == BFTW_REG) {
+		bfs_closedir(dir);
+	} else if (ftwbuf->type == BFS_REG) {
 		const struct bfs_stat *statbuf = eval_stat(state);
 		if (statbuf) {
 			ret = statbuf->size == 0;
@@ -441,6 +443,39 @@ done:
 	return ret;
 }
 
+/**
+ * -flags test.
+ */
+bool eval_flags(const struct expr *expr, struct eval_state *state) {
+	const struct bfs_stat *statbuf = eval_stat(state);
+	if (!statbuf) {
+		return false;
+	}
+
+	if (!(statbuf->mask & BFS_STAT_ATTRS)) {
+		eval_error(state, "Couldn't get file %s.\n", bfs_stat_field_name(BFS_STAT_ATTRS));
+		return false;
+	}
+
+	unsigned long flags = statbuf->attrs;
+	unsigned long set = expr->set_flags;
+	unsigned long clear = expr->clear_flags;
+
+	switch (expr->mode_cmp) {
+	case MODE_EXACT:
+		return flags == set && !(flags & clear);
+
+	case MODE_ALL:
+		return (flags & set) == set && !(flags & clear);
+
+	case MODE_ANY:
+		return (flags & set) || (flags & clear) != clear;
+	}
+
+	assert(!"Invalid comparison mode");
+	return false;
+}
+
 /**
  * -fstype test.
  */
@@ -465,7 +500,14 @@ bool eval_fstype(const struct expr *expr, struct eval_state *state) {
  */
 bool eval_hidden(const struct expr *expr, struct eval_state *state) {
 	const struct BFTW *ftwbuf = state->ftwbuf;
-	return ftwbuf->nameoff > 0 && ftwbuf->path[ftwbuf->nameoff] == '.';
+	const char *name = ftwbuf->path + ftwbuf->nameoff;
+
+	// Don't treat "." or ".." as hidden directories.  Otherwise we'd filter
+	// out everything when given
+	//
+	//     $ bfs . -nohidden
+	//     $ bfs .. -nohidden
+	return name[0] == '.' && strcmp(name, ".") != 0 && strcmp(name, "..") != 0;
 }
 
 /**
@@ -500,7 +542,7 @@ bool eval_lname(const struct expr *expr, struct eval_state *state) {
 	char *name = NULL;
 
 	const struct BFTW *ftwbuf = state->ftwbuf;
-	if (ftwbuf->type != BFTW_LNK) {
+	if (ftwbuf->type != BFS_LNK) {
 		goto done;
 	}
 
@@ -566,7 +608,7 @@ bool eval_perm(const struct expr *expr, struct eval_state *state) {
 
 	mode_t mode = statbuf->mode;
 	mode_t target;
-	if (state->ftwbuf->type == BFTW_DIR) {
+	if (state->ftwbuf->type == BFS_DIR) {
 		target = expr->dir_mode;
 	} else {
 		target = expr->file_mode;
@@ -583,6 +625,7 @@ bool eval_perm(const struct expr *expr, struct eval_state *state) {
 		return !(mode & target) == !target;
 	}
 
+	assert(!"Invalid comparison mode");
 	return false;
 }
 
@@ -634,7 +677,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		}
 	}
 
-	if (ftwbuf->type == BFTW_BLK || ftwbuf->type == BFTW_CHR) {
+	if (ftwbuf->type == BFS_BLK || ftwbuf->type == BFS_CHR) {
 		int ma = bfs_major(statbuf->rdev);
 		int mi = bfs_minor(statbuf->rdev);
 		if (fprintf(file, " %3d, %3d", ma, mi) < 0) {
@@ -672,7 +715,7 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 		goto error;
 	}
 
-	if (ftwbuf->type == BFTW_LNK) {
+	if (ftwbuf->type == BFS_LNK) {
 		if (cfprintf(cfile, " -> %pL", ftwbuf) < 0) {
 			goto error;
 		}
@@ -836,11 +879,11 @@ bool eval_size(const struct expr *expr, struct eval_state *state) {
 		[SIZE_BLOCKS] = 512,
 		[SIZE_BYTES] = 1,
 		[SIZE_WORDS] = 2,
-		[SIZE_KB] = 1024,
-		[SIZE_MB] = 1024LL*1024,
-		[SIZE_GB] = 1024LL*1024*1024,
-		[SIZE_TB] = 1024LL*1024*1024*1024,
-		[SIZE_PB] = 1024LL*1024*1024*1024*1024,
+		[SIZE_KB] = 1LL << 10,
+		[SIZE_MB] = 1LL << 20,
+		[SIZE_GB] = 1LL << 30,
+		[SIZE_TB] = 1LL << 40,
+		[SIZE_PB] = 1LL << 50,
 	};
 
 	off_t scale = scales[expr->size_unit];
@@ -900,8 +943,8 @@ bool eval_xattrname(const struct expr *expr, struct eval_state *state) {
 bool eval_xtype(const struct expr *expr, struct eval_state *state) {
 	const struct BFTW *ftwbuf = state->ftwbuf;
 	enum bfs_stat_flags flags = ftwbuf->stat_flags ^ (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW);
-	enum bftw_type type = bftw_type(ftwbuf, flags);
-	if (type == BFTW_ERROR) {
+	enum bfs_type type = bftw_type(ftwbuf, flags);
+	if (type == BFS_ERROR) {
 		eval_report_error(state);
 		return false;
 	} else {
@@ -1084,24 +1127,24 @@ static void eval_status(struct eval_state *state, struct bfs_bar *bar, struct ti
 	while (pathlen > 0) {
 		wchar_t wc;
 		size_t len = mbrtowc(&wc, path, pathlen, &mb);
-		int width;
+		int cwidth;
 		if (len == (size_t)-1) {
-			// Invalid byte sequence, assume a single-width ?
+			// Invalid byte sequence, assume a single-width '?'
 			len = 1;
-			width = 1;
+			cwidth = 1;
 			memset(&mb, 0, sizeof(mb));
 		} else if (len == (size_t)-2) {
-			// Incomplete byte sequence, assume a single-width ?
+			// Incomplete byte sequence, assume a single-width '?'
 			len = pathlen;
-			width = 1;
+			cwidth = 1;
 		} else {
-			width = wcwidth(wc);
-			if (width < 0) {
-				width = 0;
+			cwidth = wcwidth(wc);
+			if (cwidth < 0) {
+				cwidth = 0;
 			}
 		}
 
-		if (pathwidth + width > pathmax) {
+		if (pathwidth + cwidth > pathmax) {
 			break;
 		}
 
@@ -1111,7 +1154,7 @@ static void eval_status(struct eval_state *state, struct bfs_bar *bar, struct ti
 
 		path += len;
 		pathlen -= len;
-		pathwidth += width;
+		pathwidth += cwidth;
 	}
 
 	if (dstrcat(&status, "...") != 0) {
@@ -1222,28 +1265,28 @@ static void debug_stats(const struct bfs_ctx *ctx, const struct BFTW *ftwbuf) {
 	}
 }
 
-#define DUMP_BFTW_MAP(value) [value] = #value
+#define DUMP_MAP(value) [value] = #value
 
 /**
- * Dump the bftw_type for -D search.
+ * Dump the bfs_type for -D search.
  */
-static const char *dump_bftw_type(enum bftw_type type) {
+static const char *dump_bfs_type(enum bfs_type type) {
 	static const char *types[] = {
-		DUMP_BFTW_MAP(BFTW_UNKNOWN),
-		DUMP_BFTW_MAP(BFTW_BLK),
-		DUMP_BFTW_MAP(BFTW_CHR),
-		DUMP_BFTW_MAP(BFTW_DIR),
-		DUMP_BFTW_MAP(BFTW_DOOR),
-		DUMP_BFTW_MAP(BFTW_FIFO),
-		DUMP_BFTW_MAP(BFTW_LNK),
-		DUMP_BFTW_MAP(BFTW_PORT),
-		DUMP_BFTW_MAP(BFTW_REG),
-		DUMP_BFTW_MAP(BFTW_SOCK),
-		DUMP_BFTW_MAP(BFTW_WHT),
+		DUMP_MAP(BFS_UNKNOWN),
+		DUMP_MAP(BFS_BLK),
+		DUMP_MAP(BFS_CHR),
+		DUMP_MAP(BFS_DIR),
+		DUMP_MAP(BFS_DOOR),
+		DUMP_MAP(BFS_FIFO),
+		DUMP_MAP(BFS_LNK),
+		DUMP_MAP(BFS_PORT),
+		DUMP_MAP(BFS_REG),
+		DUMP_MAP(BFS_SOCK),
+		DUMP_MAP(BFS_WHT),
 	};
 
-	if (type == BFTW_ERROR) {
-		return "BFTW_ERROR";
+	if (type == BFS_ERROR) {
+		return "BFS_ERROR";
 	} else {
 		return types[type];
 	}
@@ -1254,8 +1297,8 @@ static const char *dump_bftw_type(enum bftw_type type) {
  */
 static const char *dump_bftw_visit(enum bftw_visit visit) {
 	static const char *visits[] = {
-		DUMP_BFTW_MAP(BFTW_PRE),
-		DUMP_BFTW_MAP(BFTW_POST),
+		DUMP_MAP(BFTW_PRE),
+		DUMP_MAP(BFTW_POST),
 	};
 	return visits[visit];
 }
@@ -1265,9 +1308,9 @@ static const char *dump_bftw_visit(enum bftw_visit visit) {
  */
 static const char *dump_bftw_action(enum bftw_action action) {
 	static const char *actions[] = {
-		DUMP_BFTW_MAP(BFTW_CONTINUE),
-		DUMP_BFTW_MAP(BFTW_PRUNE),
-		DUMP_BFTW_MAP(BFTW_STOP),
+		DUMP_MAP(BFTW_CONTINUE),
+		DUMP_MAP(BFTW_PRUNE),
+		DUMP_MAP(BFTW_STOP),
 	};
 	return actions[action];
 }
@@ -1313,9 +1356,8 @@ static enum bftw_action eval_callback(const struct BFTW *ftwbuf, void *ptr) {
 		eval_status(&state, args->bar, &args->last_status, args->count);
 	}
 
-	if (ftwbuf->type == BFTW_ERROR) {
+	if (ftwbuf->type == BFS_ERROR) {
 		if (!eval_should_ignore(&state, ftwbuf->error)) {
-			args->ret = EXIT_FAILURE;
 			eval_error(&state, "%s.\n", strerror(ftwbuf->error));
 		}
 		state.action = BFTW_PRUNE;
@@ -1334,7 +1376,6 @@ static enum bftw_action eval_callback(const struct BFTW *ftwbuf, void *ptr) {
 	}
 
 	if (ctx->xargs_safe && strpbrk(ftwbuf->path, " \t\n\'\"\\")) {
-		args->ret = EXIT_FAILURE;
 		eval_error(&state, "Path is not safe for xargs.\n");
 		state.action = BFTW_PRUNE;
 		goto done;
@@ -1347,7 +1388,7 @@ static enum bftw_action eval_callback(const struct BFTW *ftwbuf, void *ptr) {
 	// In -depth mode, only handle directories on the BFTW_POST visit
 	enum bftw_visit expected_visit = BFTW_PRE;
 	if ((ctx->flags & BFTW_POST_ORDER)
-	    && (ctx->strategy == BFTW_IDS || ftwbuf->type == BFTW_DIR)
+	    && (ctx->strategy == BFTW_IDS || ftwbuf->type == BFS_DIR)
 	    && ftwbuf->depth < (size_t)ctx->maxdepth) {
 		expected_visit = BFTW_POST;
 	}
@@ -1366,7 +1407,7 @@ done:
 		fprintf(stderr, "\t.root = \"%s\",\n", ftwbuf->root);
 		fprintf(stderr, "\t.depth = %zu,\n", ftwbuf->depth);
 		fprintf(stderr, "\t.visit = %s,\n", dump_bftw_visit(ftwbuf->visit));
-		fprintf(stderr, "\t.type = %s,\n", dump_bftw_type(ftwbuf->type));
+		fprintf(stderr, "\t.type = %s,\n", dump_bfs_type(ftwbuf->type));
 		fprintf(stderr, "\t.error = %d,\n", ftwbuf->error);
 		fprintf(stderr, "}) == %s\n", dump_bftw_action(state.action));
 	}
@@ -1374,41 +1415,63 @@ done:
 	return state.action;
 }
 
-/**
- * Infer the number of open file descriptors we're allowed to have.
- */
-static int infer_fdlimit(const struct bfs_ctx *ctx) {
-	int ret = 4096;
+/** Compare two rlimit values, accounting for RLIM_INFINITY etc. */
+static int rlim_cmp(rlim_t a, rlim_t b) {
+	// Consider RLIM_{INFINITY,SAVED_{CUR,MAX}} all equally infinite
+	bool a_inf = a == RLIM_INFINITY || a == RLIM_SAVED_CUR || a == RLIM_SAVED_MAX;
+	bool b_inf = b == RLIM_INFINITY || b == RLIM_SAVED_CUR || b == RLIM_SAVED_MAX;
+	if (a_inf || b_inf) {
+		return a_inf - b_inf;
+	}
 
-	struct rlimit rl;
-	if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
-		if (rl.rlim_cur != RLIM_INFINITY) {
-			ret = rl.rlim_cur;
+	return (a > b) - (a < b);
+}
+
+/** Raise RLIMIT_NOFILE if possible, and return the new limit. */
+static int raise_fdlimit(const struct bfs_ctx *ctx) {
+	rlim_t target = 64 << 10;
+	if (rlim_cmp(target, ctx->nofile_hard) > 0) {
+		target = ctx->nofile_hard;
+	}
+
+	int ret = target;
+
+	if (rlim_cmp(target, ctx->nofile_soft) > 0) {
+		const struct rlimit rl = {
+			.rlim_cur = target,
+			.rlim_max = ctx->nofile_hard,
+		};
+		if (setrlimit(RLIMIT_NOFILE, &rl) != 0) {
+			ret = ctx->nofile_soft;
 		}
 	}
 
+	return ret;
+}
+
+/** Infer the number of file descriptors available to bftw(). */
+static int infer_fdlimit(const struct bfs_ctx *ctx, int limit) {
 	// 3 for std{in,out,err}
 	int nopen = 3 + ctx->nfiles;
 
 	// Check /proc/self/fd for the current number of open fds, if possible
 	// (we may have inherited more than just the standard ones)
-	DIR *dir = opendir("/proc/self/fd");
+	struct bfs_dir *dir = bfs_opendir(AT_FDCWD, "/proc/self/fd");
 	if (!dir) {
-		dir = opendir("/dev/fd");
+		dir = bfs_opendir(AT_FDCWD, "/dev/fd");
 	}
 	if (dir) {
 		// Account for 'dir' itself
 		nopen = -1;
 
-		struct dirent *de;
-		while (xreaddir(dir, &de) == 0 && de) {
+		while (bfs_readdir(dir, NULL) > 0) {
 			++nopen;
 		}
 
-		closedir(dir);
+		bfs_closedir(dir);
 	}
 
-	ret -= nopen;
+	int ret = limit - nopen;
 	ret -= ctx->expr->persistent_fds;
 	ret -= ctx->expr->ephemeral_fds;
 
@@ -1443,10 +1506,10 @@ static void dump_bftw_flags(enum bftw_flags flags) {
  */
 static const char *dump_bftw_strategy(enum bftw_strategy strategy) {
 	static const char *strategies[] = {
-		DUMP_BFTW_MAP(BFTW_BFS),
-		DUMP_BFTW_MAP(BFTW_DFS),
-		DUMP_BFTW_MAP(BFTW_IDS),
-		DUMP_BFTW_MAP(BFTW_EDS),
+		DUMP_MAP(BFTW_BFS),
+		DUMP_MAP(BFTW_DFS),
+		DUMP_MAP(BFTW_IDS),
+		DUMP_MAP(BFTW_EDS),
 	};
 	return strategies[strategy];
 }
@@ -1474,12 +1537,15 @@ int bfs_eval(const struct bfs_ctx *ctx) {
 		args.seen = &seen;
 	}
 
+	int fdlimit = raise_fdlimit(ctx);
+	fdlimit = infer_fdlimit(ctx, fdlimit);
+
 	struct bftw_args bftw_args = {
 		.paths = ctx->paths,
 		.npaths = darray_length(ctx->paths),
 		.callback = eval_callback,
 		.ptr = &args,
-		.nopenfd = infer_fdlimit(ctx),
+		.nopenfd = fdlimit,
 		.flags = ctx->flags,
 		.strategy = ctx->strategy,
 		.mtab = bfs_ctx_mtab(ctx),
