@@ -16,7 +16,7 @@
 
 #include "exec.h"
 #include "bftw.h"
-#include "cmdline.h"
+#include "ctx.h"
 #include "color.h"
 #include "diag.h"
 #include "dstring.h"
@@ -25,19 +25,21 @@
 #include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <limits.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/types.h>
+#include <sys/resource.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
 /** Print some debugging info. */
 BFS_FORMATTER(2, 3)
 static void bfs_exec_debug(const struct bfs_exec *execbuf, const char *format, ...) {
-	if (!(execbuf->cmdline->debug & DEBUG_EXEC)) {
+	const struct bfs_ctx *ctx = execbuf->ctx;
+
+	if (!bfs_debug(ctx, DEBUG_EXEC, "${blu}")) {
 		return;
 	}
 
@@ -49,7 +51,7 @@ static void bfs_exec_debug(const struct bfs_exec *execbuf, const char *format, .
 	if (execbuf->flags & BFS_EXEC_CHDIR) {
 		fputs("dir", stderr);
 	}
-	fputs(": ", stderr);
+	cfprintf(ctx->cerr, "${rs}: ");
 
 	va_list args;
 	va_start(args, format);
@@ -65,7 +67,7 @@ static size_t bfs_exec_arg_size(const char *arg) {
 }
 
 /** Even if we can pass a bigger argument list, cap it here. */
-#define BFS_EXEC_ARG_MAX (16*1024*1024)
+#define BFS_EXEC_ARG_MAX (16 << 20)
 
 /** Determine the maximum argv size. */
 static size_t bfs_exec_arg_max(const struct bfs_exec *execbuf) {
@@ -115,15 +117,15 @@ static size_t bfs_exec_arg_max(const struct bfs_exec *execbuf) {
 	return arg_max;
 }
 
-struct bfs_exec *parse_bfs_exec(char **argv, enum bfs_exec_flags flags, const struct cmdline *cmdline) {
+struct bfs_exec *bfs_exec_parse(const struct bfs_ctx *ctx, char **argv, enum bfs_exec_flags flags) {
 	struct bfs_exec *execbuf = malloc(sizeof(*execbuf));
 	if (!execbuf) {
-		perror("malloc()");
+		bfs_perror(ctx, "malloc()");
 		goto fail;
 	}
 
 	execbuf->flags = flags;
-	execbuf->cmdline = cmdline;
+	execbuf->ctx = ctx;
 	execbuf->argv = NULL;
 	execbuf->argc = 0;
 	execbuf->argv_cap = 0;
@@ -139,9 +141,9 @@ struct bfs_exec *parse_bfs_exec(char **argv, enum bfs_exec_flags flags, const st
 		const char *arg = argv[i];
 		if (!arg) {
 			if (execbuf->flags & BFS_EXEC_CONFIRM) {
-				bfs_error(cmdline, "%s: Expected '... ;'.\n", argv[0]);
+				bfs_error(ctx, "%s: Expected '... ;'.\n", argv[0]);
 			} else {
-				bfs_error(cmdline, "%s: Expected '... ;' or '... {} +'.\n", argv[0]);
+				bfs_error(ctx, "%s: Expected '... ;' or '... {} +'.\n", argv[0]);
 			}
 			goto fail;
 		} else if (strcmp(arg, ";") == 0) {
@@ -158,14 +160,14 @@ struct bfs_exec *parse_bfs_exec(char **argv, enum bfs_exec_flags flags, const st
 	execbuf->tmpl_argc = i - 1;
 
 	if (execbuf->tmpl_argc == 0) {
-		bfs_error(cmdline, "%s: Missing command.\n", argv[0]);
+		bfs_error(ctx, "%s: Missing command.\n", argv[0]);
 		goto fail;
 	}
 
 	execbuf->argv_cap = execbuf->tmpl_argc + 1;
 	execbuf->argv = malloc(execbuf->argv_cap*sizeof(*execbuf->argv));
 	if (!execbuf->argv) {
-		perror("malloc()");
+		bfs_perror(ctx, "malloc()");
 		goto fail;
 	}
 
@@ -173,7 +175,7 @@ struct bfs_exec *parse_bfs_exec(char **argv, enum bfs_exec_flags flags, const st
 		for (i = 0; i < execbuf->tmpl_argc - 1; ++i) {
 			char *arg = execbuf->tmpl_argv[i];
 			if (strstr(arg, "{}")) {
-				bfs_error(cmdline, "%s ... +: Only one '{}' is supported.\n", argv[0]);
+				bfs_error(ctx, "%s ... +: Only one '{}' is supported.\n", argv[0]);
 				goto fail;
 			}
 			execbuf->argv[i] = arg;
@@ -186,7 +188,7 @@ struct bfs_exec *parse_bfs_exec(char **argv, enum bfs_exec_flags flags, const st
 	return execbuf;
 
 fail:
-	free_bfs_exec(execbuf);
+	bfs_exec_free(execbuf);
 	return NULL;
 }
 
@@ -357,6 +359,15 @@ static int bfs_exec_spawn(const struct bfs_exec *execbuf) {
 		goto fail;
 	}
 
+	// Reset RLIMIT_NOFILE, to avoid breaking applications that use select()
+	struct rlimit rl = {
+		.rlim_cur = execbuf->ctx->nofile_soft,
+		.rlim_max = execbuf->ctx->nofile_hard,
+	};
+	if (bfs_spawn_addsetrlimit(&ctx, RLIMIT_NOFILE, &rl) != 0) {
+		goto fail;
+	}
+
 	if (execbuf->wd_fd >= 0) {
 		if (bfs_spawn_addfchdir(&ctx, execbuf->wd_fd) != 0) {
 			goto fail;
@@ -388,9 +399,13 @@ fail:
 		}
 	} else if (WIFSIGNALED(wstatus)) {
 		int sig = WTERMSIG(wstatus);
-		bfs_warning(execbuf->cmdline, "Command '${ex}%s${rs}' terminated by signal %d\n", execbuf->argv[0], sig);
+		const char *str = strsignal(sig);
+		if (!str) {
+			str = "unknown";
+		}
+		bfs_warning(execbuf->ctx, "Command '${ex}%s${rs}' terminated by signal %d (%s)\n", execbuf->argv[0], sig, str);
 	} else {
-		bfs_warning(execbuf->cmdline, "Command '${ex}%s${rs}' terminated abnormally\n", execbuf->argv[0]);
+		bfs_warning(execbuf->ctx, "Command '${ex}%s${rs}' terminated abnormally\n", execbuf->argv[0]);
 	}
 
 	errno = 0;
@@ -607,7 +622,7 @@ int bfs_exec_finish(struct bfs_exec *execbuf) {
 	return execbuf->ret;
 }
 
-void free_bfs_exec(struct bfs_exec *execbuf) {
+void bfs_exec_free(struct bfs_exec *execbuf) {
 	if (execbuf) {
 		bfs_exec_closewd(execbuf, NULL);
 		free(execbuf->argv);
