@@ -22,6 +22,7 @@
 #define BFS_EXPR_H
 
 #include "color.h"
+#include "eval.h"
 #include "exec.h"
 #include "printf.h"
 #include "stat.h"
@@ -36,23 +37,6 @@
  */
 struct expr;
 
-/**
- * Ephemeral state for evaluating an expression.
- */
-struct eval_state;
-
-/**
- * Expression evaluation function.
- *
- * @param expr
- *         The current expression.
- * @param state
- *         The current evaluation state.
- * @return
- *         The result of the test.
- */
-typedef bool eval_fn(const struct expr *expr, struct eval_state *state);
-
 /**
  * Possible types of numeric comparison.
  */
@@ -81,6 +65,8 @@ enum mode_cmp {
  * Possible time units.
  */
 enum time_unit {
+	/** Seconds. */
+	SECONDS,
 	/** Minutes. */
 	MINUTES,
 	/** Days. */
@@ -151,6 +137,11 @@ struct expr {
 	/** Mode to use for directories (different due to X). */
 	mode_t dir_mode;
 
+	/** Flags that should be set. */
+	unsigned long long set_flags;
+	/** Flags that should be cleared. */
+	unsigned long long clear_flags;
+
 	/** The optional stat field to look at. */
 	enum bfs_stat_field stat_field;
 	/** The optional reference time. */
@@ -210,11 +201,6 @@ bool expr_never_returns(const struct expr *expr);
  */
 bool expr_cmp(const struct expr *expr, long long n);
 
-/**
- * Dump a parsed expression.
- */
-void dump_expr(CFILE *cfile, const struct expr *expr, bool verbose);
-
 /**
  * Free an expression tree.
  */
