@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -18,25 +18,27 @@
  * - main(): the entry point for bfs(1), a breadth-first version of find(1)
  *     - main.c        (this file)
  *
- * - parse_cmdline(): parses the command line into an expression tree
- *     - cmdline.h     (declares the parsed command line structure)
+ * - bfs_parse_cmdline(): parses the command line into an expression tree
+ *     - ctx.[ch]      (struct bfs_ctx, the overall bfs context)
  *     - expr.h        (declares the expression tree nodes)
- *     - parse.c       (the parser itself)
- *     - opt.c         (the expression optimizer)
+ *     - parse.[ch]    (the parser itself)
+ *     - opt.[ch]      (the optimizer)
  *
- * - eval_cmdline(): runs the expression on every file it sees
+ * - bfs_eval(): runs the expression on every file it sees
  *     - eval.[ch]     (the main evaluation functions)
  *     - exec.[ch]     (implements -exec[dir]/-ok[dir])
  *     - printf.[ch]   (implements -[f]printf)
  *
- * - bftw(): used by eval_cmdline() to walk the directory tree(s)
+ * - bftw(): used by bfs_eval() to walk the directory tree(s)
  *     - bftw.[ch]     (an extended version of nftw(3))
  *
  * - Utilities:
  *     - bfs.h         (constants about bfs itself)
+ *     - bar.[ch]      (a terminal status bar)
  *     - color.[ch]    (for pretty terminal colors)
  *     - darray.[ch]   (a dynamic array library)
  *     - diag.[ch]     (formats diagnostic messages)
+ *     - dir.[ch]      (a directory API facade)
  *     - dstring.[ch]  (a dynamic string library)
  *     - fsade.[ch]    (a facade over non-standard filesystem features)
  *     - mtab.[ch]     (parses the system's mount table)
@@ -49,15 +51,40 @@
  *     - util.[ch]     (everything else)
  */
 
-#include "cmdline.h"
-#include "util.h"
+#include "ctx.h"
+#include "eval.h"
+#include "parse.h"
 #include <errno.h>
 #include <fcntl.h>
 #include <locale.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
 
+/**
+ * Check if a file descriptor is open.
+ */
+static bool isopen(int fd) {
+	return fcntl(fd, F_GETFD) >= 0 || errno != EBADF;
+}
+
+/**
+ * Open a file and redirect it to a particular descriptor.
+ */
+static int redirect(int fd, const char *path, int flags) {
+	int newfd = open(path, flags);
+	if (newfd < 0 || newfd == fd) {
+		return newfd;
+	}
+
+	int ret = dup2(newfd, fd);
+	int err = errno;
+	close(newfd);
+	errno = err;
+	return ret;
+}
+
 /**
  * Make sure the standard streams std{in,out,err} are open.  If they are not,
  * future open() calls may use those file descriptors, and std{in,out,err} will
@@ -100,12 +127,12 @@ int main(int argc, char *argv[]) {
 	// Use the system locale instead of "C"
 	setlocale(LC_ALL, "");
 
-	struct cmdline *cmdline = parse_cmdline(argc, argv);
-	if (cmdline) {
-		ret = eval_cmdline(cmdline);
+	struct bfs_ctx *ctx = bfs_parse_cmdline(argc, argv);
+	if (ctx) {
+		ret = bfs_eval(ctx);
 	}
 
-	if (free_cmdline(cmdline) != 0 && ret == EXIT_SUCCESS) {
+	if (bfs_ctx_free(ctx) != 0 && ret == EXIT_SUCCESS) {
 		ret = EXIT_FAILURE;
 	}
 
