@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -23,15 +23,16 @@
 
 #include "parse.h"
 #include "bfs.h"
+#include "bftw.h"
+#include "color.h"
 #include "ctx.h"
 #include "darray.h"
 #include "diag.h"
-#include "dstring.h"
+#include "dir.h"
 #include "eval.h"
 #include "exec.h"
 #include "expr.h"
 #include "fsade.h"
-#include "mtab.h"
 #include "opt.h"
 #include "printf.h"
 #include "pwcache.h"
@@ -41,7 +42,6 @@
 #include "typo.h"
 #include "util.h"
 #include <assert.h>
-#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <fnmatch.h>
@@ -50,11 +50,11 @@
 #include <pwd.h>
 #include <regex.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <time.h>
@@ -498,7 +498,7 @@ static const char *parse_int(const struct parser_state *state, const char *str,
 		break;
 
 	default:
-		assert(false);
+		assert(!"Invalid int size");
 		goto bad;
 	}
 
@@ -737,25 +737,6 @@ static void debug_help(CFILE *cfile) {
 	cfprintf(cfile, "  ${bld}all${rs}:    All debug flags at once.\n");
 }
 
-/** A named debug flag. */
-struct debug_flag {
-	enum debug_flags flag;
-	const char *name;
-};
-
-/** The table of debug flags. */
-struct debug_flag debug_flags[] = {
-	{DEBUG_ALL,    "all"},
-	{DEBUG_COST,   "cost"},
-	{DEBUG_EXEC,   "exec"},
-	{DEBUG_OPT,    "opt"},
-	{DEBUG_RATES,  "rates"},
-	{DEBUG_SEARCH, "search"},
-	{DEBUG_STAT,   "stat"},
-	{DEBUG_TREE,   "tree"},
-	{0},
-};
-
 /** Check if a substring matches a debug flag. */
 static bool parse_debug_flag(const char *flag, size_t len, const char *expected) {
 	if (len == strlen(expected)) {
@@ -793,22 +774,26 @@ static struct expr *parse_debug(struct parser_state *state, int arg1, int arg2)
 			debug_help(ctx->cout);
 			state->just_info = true;
 			return NULL;
+		} else if (parse_debug_flag(flag, len, "all")) {
+			ctx->debug = DEBUG_ALL;
+			continue;
 		}
 
-		for (int i = 0; ; ++i) {
-			const char *expected = debug_flags[i].name;
-			if (!expected) {
-				if (parse_warning(state, "Unrecognized debug flag ${bld}")) {
-					fwrite(flag, 1, len, stderr);
-					cfprintf(ctx->cerr, "${rs}.\n\n");
-					unrecognized = true;
-				}
+		enum debug_flags i;
+		for (i = 1; DEBUG_ALL & i; i <<= 1) {
+			const char *name = debug_flag_name(i);
+			if (parse_debug_flag(flag, len, name)) {
 				break;
 			}
+		}
 
-			if (parse_debug_flag(flag, len, expected)) {
-				ctx->debug |= debug_flags[i].flag;
-				break;
+		if (DEBUG_ALL & i) {
+			ctx->debug |= i;
+		} else {
+			if (parse_warning(state, "Unrecognized debug flag ${bld}")) {
+				fwrite(flag, 1, len, stderr);
+				cfprintf(ctx->cerr, "${rs}.\n\n");
+				unrecognized = true;
 			}
 		}
 	}
@@ -931,9 +916,9 @@ fail:
 }
 
 /**
- * Parse -[aBcm]{min,time}.
+ * Parse -[aBcm]min.
  */
-static struct expr *parse_time(struct parser_state *state, int field, int unit) {
+static struct expr *parse_min(struct parser_state *state, int field, int arg2) {
 	struct expr *expr = parse_test_icmp(state, eval_time);
 	if (!expr) {
 		return NULL;
@@ -942,10 +927,83 @@ static struct expr *parse_time(struct parser_state *state, int field, int unit)
 	expr->cost = STAT_COST;
 	expr->reftime = state->now;
 	expr->stat_field = field;
-	expr->time_unit = unit;
+	expr->time_unit = MINUTES;
 	return expr;
 }
 
+/**
+ * Parse -[aBcm]time.
+ */
+static struct expr *parse_time(struct parser_state *state, int field, int arg2) {
+	struct expr *expr = parse_unary_test(state, eval_time);
+	if (!expr) {
+		return NULL;
+	}
+
+	expr->cost = STAT_COST;
+	expr->reftime = state->now;
+	expr->stat_field = field;
+
+	const char *tail = parse_icmp(state, expr->sdata, expr, IF_PARTIAL_OK);
+	if (!tail) {
+		goto fail;
+	}
+
+	if (!*tail) {
+		expr->time_unit = DAYS;
+		return expr;
+	}
+
+	unsigned long long time = expr->idata;
+	expr->idata = 0;
+
+	while (true) {
+		switch (*tail) {
+		case 'w':
+			time *= 7;
+			fallthrough;
+		case 'd':
+			time *= 24;
+			fallthrough;
+		case 'h':
+			time *= 60;
+			fallthrough;
+		case 'm':
+			time *= 60;
+			fallthrough;
+		case 's':
+			break;
+		default:
+			parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: Unknown time unit ${bld}%c${rs}.\n",
+			            expr->argv[0], expr->argv[1], *tail);
+			goto fail;
+		}
+
+		expr->idata += time;
+
+		if (!*++tail) {
+			break;
+		}
+
+		tail = parse_int(state, tail, &time, IF_PARTIAL_OK | IF_LONG_LONG | IF_UNSIGNED);
+		if (!tail) {
+			goto fail;
+		}
+		if (!*tail) {
+			parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: Missing time unit.\n",
+			            expr->argv[0], expr->argv[1]);
+			goto fail;
+		}
+	}
+
+	expr->time_unit = SECONDS;
+	return expr;
+
+fail:
+	free_expr(expr);
+	return NULL;
+}
+
 /**
  * Parse -capable.
  */
@@ -1174,6 +1232,43 @@ static struct expr *parse_f(struct parser_state *state, int arg1, int arg2) {
 	return &expr_true;
 }
 
+/**
+ * Parse -flags FLAGS.
+ */
+static struct expr *parse_flags(struct parser_state *state, int arg1, int arg2) {
+	struct expr *expr = parse_unary_test(state, eval_flags);
+	if (!expr) {
+		return NULL;
+	}
+
+	const char *flags = expr->sdata;
+	switch (flags[0]) {
+	case '-':
+		expr->mode_cmp = MODE_ALL;
+		++flags;
+		break;
+	case '+':
+		expr->mode_cmp = MODE_ANY;
+		++flags;
+		break;
+	default:
+		expr->mode_cmp = MODE_EXACT;
+		break;
+	}
+
+	if (xstrtofflags(&flags, &expr->set_flags, &expr->clear_flags) != 0) {
+		if (errno == ENOTSUP) {
+			parse_error(state, "${blu}%s${rs} is missing platform support.\n", expr->argv[0]);
+		} else {
+			parse_error(state, "${blu}%s${rs}: Invalid flags ${bld}%s${rs}.\n", expr->argv[0], flags);
+		}
+		free_expr(expr);
+		return NULL;
+	}
+
+	return expr;
+}
+
 /**
  * Parse -fls FILE.
  */
@@ -1778,7 +1873,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 		case MODE_CLAUSE:
 			who = 0;
 			mstate = MODE_WHO;
-			// Fallthrough
+			fallthrough;
 
 		case MODE_WHO:
 			switch (*i) {
@@ -1805,7 +1900,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 			case MODE_EQUALS:
 				expr->file_mode &= ~who;
 				expr->dir_mode &= ~who;
-				// Fallthrough
+				fallthrough;
 			case MODE_PLUS:
 				expr->file_mode |= file_change;
 				expr->dir_mode |= dir_change;
@@ -1815,7 +1910,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				expr->dir_mode &= ~dir_change;
 				break;
 			}
-			// Fallthrough
+			fallthrough;
 
 		case MODE_ACTION:
 			if (who == 0) {
@@ -1897,7 +1992,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				break;
 			case 'x':
 				file_change |= who & 0111;
-				// Fallthrough
+				fallthrough;
 			case 'X':
 				dir_change |= who & 0111;
 				break;
@@ -1912,8 +2007,10 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				}
 				break;
 			case 't':
-				file_change |= S_ISVTX;
-				dir_change |= S_ISVTX;
+				if (who & 0007) {
+					file_change |= S_ISVTX;
+					dir_change |= S_ISVTX;
+				}
 				break;
 			default:
 				mstate = MODE_ACTION_APPLY;
@@ -1958,7 +2055,7 @@ static struct expr *parse_perm(struct parser_state *state, int field, int arg2)
 			++mode;
 			break;
 		}
-		// Fallthrough
+		fallthrough;
 	default:
 		expr->mode_cmp = MODE_EXACT;
 		break;
@@ -2071,7 +2168,7 @@ static struct expr *parse_regex(struct parser_state *state, int flags, int arg2)
 
 	int err = regcomp(expr->regex, expr->sdata, state->regex_flags | flags);
 	if (err != 0) {
-		char *str = xregerror(err, expr->regex);
+		char *str = xregerror(err, NULL);
 		if (str) {
 			parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: %s.\n", expr->argv[0], expr->argv[1], str);
 			free(str);
@@ -2328,44 +2425,44 @@ static struct expr *parse_type(struct parser_state *state, int x, int arg2) {
 
 	const char *c = expr->sdata;
 	while (true) {
-		enum bftw_type type;
+		enum bfs_type type;
 		double type_prob;
 
 		switch (*c) {
 		case 'b':
-			type = BFTW_BLK;
+			type = BFS_BLK;
 			type_prob = 0.00000721183;
 			break;
 		case 'c':
-			type = BFTW_CHR;
+			type = BFS_CHR;
 			type_prob = 0.0000499855;
 			break;
 		case 'd':
-			type = BFTW_DIR;
+			type = BFS_DIR;
 			type_prob = 0.114475;
 			break;
 		case 'D':
-			type = BFTW_DOOR;
+			type = BFS_DOOR;
 			type_prob = 0.000001;
 			break;
 		case 'p':
-			type = BFTW_FIFO;
+			type = BFS_FIFO;
 			type_prob = 0.00000248684;
 			break;
 		case 'f':
-			type = BFTW_REG;
+			type = BFS_REG;
 			type_prob = 0.859772;
 			break;
 		case 'l':
-			type = BFTW_LNK;
+			type = BFS_LNK;
 			type_prob = 0.0256816;
 			break;
 		case 's':
-			type = BFTW_SOCK;
+			type = BFS_SOCK;
 			type_prob = 0.0000116881;
 			break;
 		case 'w':
-			type = BFTW_WHT;
+			type = BFS_WHT;
 			type_prob = 0.000001;
 			break;
 
@@ -2470,7 +2567,7 @@ static struct expr *parse_xdev(struct parser_state *state, int arg1, int arg2) {
  */
 static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
 	char *pager = getenv("PAGER");
-	if (!pager) {
+	if (!pager || !pager[0]) {
 		pager = "more";
 	}
 
@@ -2520,19 +2617,24 @@ static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
 	extern char **environ;
 	char **envp = environ;
 
-	if (!getenv("LESS")) {
+	const char *less = getenv("LESS");
+	if (!less || !less[0]) {
 		size_t envc;
-		for (envc = 0; environ[envc]; ++envc);
-		++envc;
+		for (envc = 0; environ[envc]; ++envc) { }
 
-		envp = malloc((envc + 1)*sizeof(*envp));
+		envp = malloc((envc + 2)*sizeof(*envp));
 		if (!envp) {
 			goto fail_ctx;
 		}
 
-		memcpy(envp, environ, (envc - 1)*sizeof(*envp));
-		envp[envc - 1] = "LESS=FKRX";
-		envp[envc] = NULL;
+		size_t j = 0;
+		for (size_t i = 0; i < envc; ++i) {
+			if (strncmp(environ[i], "LESS=", 5) != 0) {
+				envp[j++] = environ[i];
+			}
+		}
+		envp[j++] = "LESS=FKRX";
+		envp[j] = NULL;
 	}
 
 	*pid = bfs_spawn(pager, &ctx, argv, envp);
@@ -2864,10 +2966,10 @@ static const struct table_entry parse_table[] = {
 	{"--", T_FLAG},
 	{"--help", T_ACTION, parse_help},
 	{"--version", T_ACTION, parse_version},
-	{"-Bmin", T_TEST, parse_time, BFS_STAT_BTIME, MINUTES},
+	{"-Bmin", T_TEST, parse_min, BFS_STAT_BTIME},
 	{"-Bnewer", T_TEST, parse_newer, BFS_STAT_BTIME},
 	{"-Bsince", T_TEST, parse_since, BFS_STAT_BTIME},
-	{"-Btime", T_TEST, parse_time, BFS_STAT_BTIME, DAYS},
+	{"-Btime", T_TEST, parse_time, BFS_STAT_BTIME},
 	{"-D", T_FLAG, parse_debug},
 	{"-E", T_FLAG, parse_regex_extended},
 	{"-H", T_FLAG, parse_follow, BFTW_FOLLOW_ROOTS, false},
@@ -2878,17 +2980,17 @@ static const struct table_entry parse_table[] = {
 	{"-X", T_FLAG, parse_xargs_safe},
 	{"-a", T_OPERATOR},
 	{"-acl", T_TEST, parse_acl},
-	{"-amin", T_TEST, parse_time, BFS_STAT_ATIME, MINUTES},
+	{"-amin", T_TEST, parse_min, BFS_STAT_ATIME},
 	{"-and", T_OPERATOR},
 	{"-anewer", T_TEST, parse_newer, BFS_STAT_ATIME},
 	{"-asince", T_TEST, parse_since, BFS_STAT_ATIME},
-	{"-atime", T_TEST, parse_time, BFS_STAT_ATIME, DAYS},
+	{"-atime", T_TEST, parse_time, BFS_STAT_ATIME},
 	{"-capable", T_TEST, parse_capable},
-	{"-cmin", T_TEST, parse_time, BFS_STAT_CTIME, MINUTES},
+	{"-cmin", T_TEST, parse_min, BFS_STAT_CTIME},
 	{"-cnewer", T_TEST, parse_newer, BFS_STAT_CTIME},
 	{"-color", T_OPTION, parse_color, true},
 	{"-csince", T_TEST, parse_since, BFS_STAT_CTIME},
-	{"-ctime", T_TEST, parse_time, BFS_STAT_CTIME, DAYS},
+	{"-ctime", T_TEST, parse_time, BFS_STAT_CTIME},
 	{"-d", T_FLAG, parse_depth},
 	{"-daystart", T_OPTION, parse_daystart},
 	{"-delete", T_ACTION, parse_delete},
@@ -2901,6 +3003,7 @@ static const struct table_entry parse_table[] = {
 	{"-exit", T_ACTION, parse_exit},
 	{"-f", T_FLAG, parse_f},
 	{"-false", T_TEST, parse_const, false},
+	{"-flags", T_TEST, parse_flags},
 	{"-fls", T_ACTION, parse_fls},
 	{"-follow", T_OPTION, parse_follow, BFTW_FOLLOW_ALL, true},
 	{"-fprint", T_ACTION, parse_fprint},
@@ -2923,11 +3026,11 @@ static const struct table_entry parse_table[] = {
 	{"-ls", T_ACTION, parse_ls},
 	{"-maxdepth", T_OPTION, parse_depth_limit, false},
 	{"-mindepth", T_OPTION, parse_depth_limit, true},
-	{"-mmin", T_TEST, parse_time, BFS_STAT_MTIME, MINUTES},
+	{"-mmin", T_TEST, parse_min, BFS_STAT_MTIME},
 	{"-mnewer", T_TEST, parse_newer, BFS_STAT_MTIME},
 	{"-mount", T_OPTION, parse_mount},
 	{"-msince", T_TEST, parse_since, BFS_STAT_MTIME},
-	{"-mtime", T_TEST, parse_time, BFS_STAT_MTIME, DAYS},
+	{"-mtime", T_TEST, parse_time, BFS_STAT_MTIME},
 	{"-name", T_TEST, parse_name, false},
 	{"-newer", T_TEST, parse_newer, BFS_STAT_MTIME},
 	{"-newer", T_TEST, parse_newerxy, 0, 0, true},
@@ -2938,7 +3041,7 @@ static const struct table_entry parse_table[] = {
 	{"-noleaf", T_OPTION, parse_noleaf},
 	{"-not", T_OPERATOR},
 	{"-nouser", T_TEST, parse_nouser},
-	{"-nowarn", T_TEST, parse_warn, false},
+	{"-nowarn", T_OPTION, parse_warn, false},
 	{"-o", T_OPERATOR},
 	{"-ok", T_ACTION, parse_exec, BFS_EXEC_CONFIRM},
 	{"-okdir", T_ACTION, parse_exec, BFS_EXEC_CONFIRM | BFS_EXEC_CHDIR},
@@ -3375,14 +3478,14 @@ void bfs_ctx_dump(const struct bfs_ctx *ctx, enum debug_flags flag) {
 	cfprintf(cerr, "${cyn}-S${rs} ${bld}%s${rs} ", strategy);
 
 	enum debug_flags debug = ctx->debug;
-	if (debug) {
+	if (debug == DEBUG_ALL) {
+		cfprintf(cerr, "${cyn}-D${rs} ${bld}all${rs} ");
+	} else if (debug) {
 		cfprintf(cerr, "${cyn}-D${rs} ");
-		for (int i = 0; debug; ++i) {
-			enum debug_flags flag = debug_flags[i].flag;
-			const char *name = debug_flags[i].name;
-			if ((debug & flag) == flag) {
-				cfprintf(cerr, "${bld}%s${rs}", name);
-				debug ^= flag;
+		for (enum debug_flags i = 1; DEBUG_ALL & i; i <<= 1) {
+			if (debug & i) {
+				cfprintf(cerr, "${bld}%s${rs}", debug_flag_name(i));
+				debug ^= i;
 				if (debug) {
 					cfprintf(cerr, ",");
 				}
