@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2017-2020 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2017-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -15,11 +15,11 @@
  ****************************************************************************/
 
 #include "printf.h"
-#include "color.h"
+#include "bftw.h"
 #include "ctx.h"
 #include "diag.h"
+#include "dir.h"
 #include "dstring.h"
-#include "expr.h"
 #include "mtab.h"
 #include "pwcache.h"
 #include "stat.h"
@@ -29,9 +29,10 @@
 #include <errno.h>
 #include <grp.h>
 #include <pwd.h>
-#include <stdlib.h>
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <time.h>
 
@@ -309,7 +310,7 @@ static int bfs_printf_l(FILE *file, const struct bfs_printf *directive, const st
 	char *buf = NULL;
 	const char *target = "";
 
-	if (ftwbuf->type == BFTW_LNK) {
+	if (ftwbuf->type == BFS_LNK) {
 		const struct bfs_stat *statbuf = bftw_cached_stat(ftwbuf, BFS_STAT_NOFOLLOW);
 		size_t len = statbuf ? statbuf->size : 0;
 
@@ -425,23 +426,23 @@ static int bfs_printf_u(FILE *file, const struct bfs_printf *directive, const st
 	return fprintf(file, directive->str, pwd->pw_name);
 }
 
-static const char *bfs_printf_type(enum bftw_type type) {
+static const char *bfs_printf_type(enum bfs_type type) {
 	switch (type) {
-	case BFTW_BLK:
+	case BFS_BLK:
 		return "b";
-	case BFTW_CHR:
+	case BFS_CHR:
 		return "c";
-	case BFTW_DIR:
+	case BFS_DIR:
 		return "d";
-	case BFTW_DOOR:
+	case BFS_DOOR:
 		return "D";
-	case BFTW_FIFO:
+	case BFS_FIFO:
 		return "p";
-	case BFTW_LNK:
+	case BFS_LNK:
 		return "l";
-	case BFTW_REG:
+	case BFS_REG:
 		return "f";
-	case BFTW_SOCK:
+	case BFS_SOCK:
 		return "s";
 	default:
 		return "U";
@@ -458,7 +459,7 @@ static int bfs_printf_y(FILE *file, const struct bfs_printf *directive, const st
 static int bfs_printf_Y(FILE *file, const struct bfs_printf *directive, const struct BFTW *ftwbuf) {
 	int error = 0;
 
-	if (ftwbuf->type != BFTW_LNK) {
+	if (ftwbuf->type != BFS_LNK) {
 		return bfs_printf_y(file, directive, ftwbuf);
 	}
 
@@ -466,7 +467,7 @@ static int bfs_printf_Y(FILE *file, const struct bfs_printf *directive, const st
 
 	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, BFS_STAT_FOLLOW);
 	if (statbuf) {
-		type = bfs_printf_type(bftw_mode_to_type(statbuf->mode));
+		type = bfs_printf_type(bfs_mode_to_type(statbuf->mode));
 	} else {
 		switch (errno) {
 		case ELOOP:
@@ -629,7 +630,7 @@ struct bfs_printf *bfs_printf_parse(const struct bfs_ctx *ctx, const char *forma
 				case '0':
 				case '+':
 					must_be_numeric = true;
-					// Fallthrough
+					fallthrough;
 				case ' ':
 				case '-':
 					if (strchr(directive->str, c)) {
