@@ -17,9 +17,7 @@
 #include "spawn.h"
 #include "util.h"
 #include <errno.h>
-#include <fcntl.h>
 #include <stdlib.h>
-#include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -31,6 +29,7 @@ enum bfs_spawn_op {
 	BFS_SPAWN_CLOSE,
 	BFS_SPAWN_DUP2,
 	BFS_SPAWN_FCHDIR,
+	BFS_SPAWN_SETRLIMIT,
 };
 
 /**
@@ -42,6 +41,8 @@ struct bfs_spawn_action {
 	enum bfs_spawn_op op;
 	int in_fd;
 	int out_fd;
+	int resource;
+	struct rlimit rlimit;
 };
 
 int bfs_spawn_init(struct bfs_spawn *ctx) {
@@ -127,6 +128,17 @@ int bfs_spawn_addfchdir(struct bfs_spawn *ctx, int fd) {
 	}
 }
 
+int bfs_spawn_addsetrlimit(struct bfs_spawn *ctx, int resource, const struct rlimit *rl) {
+	struct bfs_spawn_action *action = bfs_spawn_add(ctx, BFS_SPAWN_SETRLIMIT);
+	if (action) {
+		action->resource = resource;
+		action->rlimit = *rl;
+		return 0;
+	} else {
+		return -1;
+	}
+}
+
 /** Facade for execvpe() which is non-standard. */
 static int bfs_execvpe(const char *exe, char **argv, char **envp) {
 #if __GLIBC__ || __linux__ || __NetBSD__ || __OpenBSD__
@@ -179,6 +191,11 @@ static void bfs_spawn_exec(const char *exe, const struct bfs_spawn *ctx, char **
 				goto fail;
 			}
 			break;
+		case BFS_SPAWN_SETRLIMIT:
+			if (setrlimit(action->resource, &action->rlimit) != 0) {
+				goto fail;
+			}
+			break;
 		}
 	}
 
@@ -191,12 +208,9 @@ static void bfs_spawn_exec(const char *exe, const struct bfs_spawn *ctx, char **
 fail:
 	error = errno;
 
-	while (write(pipefd[1], &error, sizeof(error)) < 0) {
-		if (errno != EINTR) {
-			// Parent will still see that we exited unsuccessfully, but won't know why
-			break;
-		}
-	}
+	// In case of a write error, the parent will still see that we exited
+	// unsuccessfully, but won't know why
+	(void) xwrite(pipefd[1], &error, sizeof(error));
 
 	close(pipefd[1]);
 	_Exit(127);
@@ -226,7 +240,7 @@ pid_t bfs_spawn(const char *exe, const struct bfs_spawn *ctx, char **argv, char
 	// Parent
 	close(pipefd[1]);
 
-	ssize_t nbytes = read(pipefd[0], &error, sizeof(error));
+	ssize_t nbytes = xread(pipefd[0], &error, sizeof(error));
 	close(pipefd[0]);
 	if (nbytes == sizeof(error)) {
 		int wstatus;
