@@ -23,13 +23,17 @@
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <unistd.h>
+
+#if BFS_HAS_SYS_PARAM
+#	include <sys/param.h>
+#endif
 
 #ifdef STATX_BASIC_STATS
 #	define HAVE_STATX true
 #elif __linux__
 #	include <linux/stat.h>
 #	include <sys/syscall.h>
+#	include <unistd.h>
 #endif
 
 #if HAVE_STATX || defined(__NR_statx)
@@ -65,6 +69,8 @@ const char *bfs_stat_field_name(enum bfs_stat_field field) {
 		return "block count";
 	case BFS_STAT_RDEV:
 		return "underlying device";
+	case BFS_STAT_ATTRS:
+		return "attributes";
 	case BFS_STAT_ATIME:
 		return "access time";
 	case BFS_STAT_BTIME:
@@ -75,7 +81,7 @@ const char *bfs_stat_field_name(enum bfs_stat_field field) {
 		return "modification time";
 	}
 
-	assert(false);
+	assert(!"Unrecognized stat field");
 	return "???";
 }
 
@@ -121,6 +127,11 @@ static void bfs_stat_convert(const struct stat *statbuf, struct bfs_stat *buf) {
 	buf->rdev = statbuf->st_rdev;
 	buf->mask |= BFS_STAT_RDEV;
 
+#if BSD
+	buf->attrs = statbuf->st_flags;
+	buf->mask |= BFS_STAT_ATTRS;
+#endif
+
 	buf->atime = statbuf->st_atim;
 	buf->mask |= BFS_STAT_ATIME;
 
@@ -161,9 +172,9 @@ static int bfs_stat_impl(int at_fd, const char *at_path, int at_flags, enum bfs_
  * Wrapper for the statx() system call, which had no glibc wrapper prior to 2.28.
  */
 static int bfs_statx(int at_fd, const char *at_path, int at_flags, unsigned int mask, struct statx *buf) {
+#if BFS_HAS_FEATURE(memory_sanitizer, false)
 	// -fsanitize=memory doesn't know about statx(), so tell it the memory
 	// got initialized
-#if BFS_HAS_FEATURE(memory_sanitizer, false)
 	memset(buf, 0, sizeof(*buf));
 #endif
 
@@ -244,6 +255,9 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, enum bfs
 	buf->rdev = bfs_makedev(xbuf.stx_rdev_major, xbuf.stx_rdev_minor);
 	buf->mask |= BFS_STAT_RDEV;
 
+	buf->attrs = xbuf.stx_attributes;
+	buf->mask |= BFS_STAT_ATTRS;
+
 	if (xbuf.stx_mask & STATX_ATIME) {
 		buf->atime.tv_sec = xbuf.stx_atime.tv_sec;
 		buf->atime.tv_nsec = xbuf.stx_atime.tv_nsec;
@@ -349,7 +363,7 @@ const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_f
 	case BFS_STAT_MTIME:
 		return &buf->mtime;
 	default:
-		assert(false);
+		assert(!"Invalid stat field for time");
 		errno = EINVAL;
 		return NULL;
 	}
