@@ -2,7 +2,7 @@
 
 ############################################################################
 # bfs                                                                      #
-# Copyright (C) 2015-2020 Tavian Barnes <tavianator@tavianator.com>        #
+# Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        #
 #                                                                          #
 # Permission to use, copy, modify, and/or distribute this software for any #
 # purpose with or without fee is hereby granted.                           #
@@ -34,10 +34,36 @@ if [ -t 1 ]; then
     RST="$(printf '\033[0m')"
 fi
 
-if [ "$EUID" -eq 0 ]; then
+if command -v capsh &>/dev/null; then
+    if capsh --has-p=cap_dac_override &>/dev/null || capsh --has-p=cap_dac_read_search &>/dev/null; then
+	if [ -n "$BFS_TRIED_DROP" ]; then
+            cat >&2 <<EOF
+${RED}error:${RST} Failed to drop capabilities.
+EOF
+
+	    exit 1
+	fi
+
+        cat >&2 <<EOF
+${YLW}warning:${RST} Running as ${BLD}$(id -un)${RST} is not recommended.  Dropping ${BLD}cap_dac_override${RST} and
+${BLD}cap_dac_read_search${RST}.
+
+EOF
+
+        BFS_TRIED_DROP=y exec capsh \
+            --drop=cap_dac_override,cap_dac_read_search \
+            --caps=cap_dac_override,cap_dac_read_search-eip \
+            -- "$0" "$@"
+    fi
+elif [ "$EUID" -eq 0 ]; then
+    UNLESS=
+    if [ "$(uname)" = "Linux" ]; then
+	UNLESS=" unless ${GRN}capsh${RST} is installed"
+    fi
+
     cat >&2 <<EOF
 ${RED}error:${RST} These tests expect filesystem permissions to be enforced, and therefore
-will not work when run as ${BLD}$(id -un)${RST}.
+will not work when run as ${BLD}$(id -un)${RST}${UNLESS}.
 EOF
     exit 1
 fi
@@ -248,6 +274,9 @@ posix_tests=(
 
     test_type_d
     test_type_f
+    test_type_l
+    test_H_type_l
+    test_L_type_l
 
     test_user_name
     test_user_id
@@ -302,6 +331,7 @@ bsd_tests=(
     test_asince
 
     test_delete
+    test_delete_many
 
     test_depth_maxdepth_1
     test_depth_maxdepth_2
@@ -327,6 +357,8 @@ bsd_tests=(
 
     test_exit
 
+    test_flags
+
     test_follow
 
     test_gid_name
@@ -357,6 +389,8 @@ bsd_tests=(
 
     test_msince
 
+    test_mtime_units
+
     test_name_slash
     test_name_slashes
 
@@ -406,6 +440,9 @@ bsd_tests=(
     test_size_big
 
     test_uid_name
+
+    # Optimizer tests
+    test_data_flow_sparse
 )
 
 gnu_tests=(
@@ -439,6 +476,7 @@ gnu_tests=(
     test_daystart_twice
 
     test_delete
+    test_delete_many
     test_L_delete
 
     test_depth_mindepth_1
@@ -603,7 +641,6 @@ gnu_tests=(
     test_and_false_or_true
     test_comma_redundant_true
     test_comma_redundant_false
-    test_data_flow_sparse
 )
 
 bfs_tests=(
@@ -681,6 +718,7 @@ bfs_tests=(
     test_help
 
     test_hidden
+    test_hidden_root
 
     test_newerma_nonexistent
     test_newermt_invalid
@@ -855,6 +893,7 @@ function make_links() {
     mkdir -p "$1/deeply/nested/dir"
     touchp "$1/deeply/nested/file"
     ln -s file "$1/deeply/nested/link"
+    ln -s nowhere "$1/deeply/nested/broken"
     ln -s deeply/nested "$1/skip"
 }
 make_links "$TMP/links"
@@ -1093,6 +1132,18 @@ function test_type_f() {
     bfs_diff basic -type f
 }
 
+function test_type_l() {
+    bfs_diff links/skip -type l
+}
+
+function test_H_type_l() {
+    bfs_diff -H links/skip -type l
+}
+
+function test_L_type_l() {
+    bfs_diff -L links/skip -type l
+}
+
 function test_type_multi() {
     bfs_diff links -type f,d,c
 }
@@ -1204,11 +1255,15 @@ function test_gid() {
 }
 
 function test_gid_plus() {
-    bfs_diff basic -gid +0
+    if [ "$(id -g)" -ne 0 ]; then
+	bfs_diff basic -gid +0
+    fi
 }
 
 function test_gid_plus_plus() {
-    bfs_diff basic -gid +0
+    if [ "$(id -g)" -ne 0 ]; then
+	bfs_diff basic -gid ++0
+    fi
 }
 
 function test_gid_minus() {
@@ -1224,11 +1279,15 @@ function test_uid() {
 }
 
 function test_uid_plus() {
-    bfs_diff basic -uid +0
+    if [ "$(id -u)" -ne 0 ]; then
+	bfs_diff basic -uid +0
+    fi
 }
 
 function test_uid_plus_plus() {
-    bfs_diff basic -uid ++0
+    if [ "$(id -u)" -ne 0 ]; then
+	bfs_diff basic -uid ++0
+    fi
 }
 
 function test_uid_minus() {
@@ -1757,11 +1816,11 @@ function test_permcopy() {
 }
 
 function test_perm_setid() {
-    bfs_diff rainbow -perm /ug+s
+    bfs_diff rainbow -perm -u+s -o -perm -g+s
 }
 
 function test_perm_sticky() {
-    bfs_diff rainbow -perm /ug+t
+    bfs_diff rainbow -perm -a+t
 }
 
 function test_prune() {
@@ -1810,6 +1869,17 @@ function test_delete() {
     bfs_diff scratch
 }
 
+function test_delete_many() {
+    # Test for https://github.com/tavianator/bfs/issues/67
+
+    rm -rf scratch/*
+    mkdir scratch/foo
+    touch scratch/foo/{1..256}
+
+    invoke_bfs scratch/foo -delete
+    bfs_diff scratch
+}
+
 function test_L_delete() {
     rm -rf scratch/*
     mkdir scratch/foo
@@ -1890,6 +1960,11 @@ function test_hidden() {
     bfs_diff weirdnames -hidden
 }
 
+function test_hidden_root() {
+    cd weirdnames
+    bfs_diff . ./. ... ./... .../.. -hidden
+}
+
 function test_nohidden() {
     bfs_diff weirdnames -nohidden
 }
@@ -1946,6 +2021,10 @@ function test_msince() {
     bfs_diff times -msince 1991-12-14T00:01
 }
 
+function test_mtime_units() {
+    bfs_diff times -mtime +500w400d300h200m100s
+}
+
 function test_size_T() {
     bfs_diff basic -type f -size 1T
 }
@@ -2106,7 +2185,7 @@ function test_printf_u_g_ulimit() {
 }
 
 function test_printf_l_nonlink() {
-    bfs_diff links -printf '| %24p -> %-24l |\n'
+    bfs_diff links -printf '| %26p -> %-26l |\n'
 }
 
 function test_printf_incomplete_escape() {
@@ -2763,49 +2842,40 @@ function make_xattrs() {
 
     case "$UNAME" in
         Darwin)
-            xattr -w bfs_test true scratch/xattr
-            xattr -w bfs_test_2 true scratch/xattr_2
-            xattr -s -w bfs_test true scratch/xattr_link
+            xattr -w bfs_test true scratch/xattr \
+                && xattr -w bfs_test_2 true scratch/xattr_2 \
+                && xattr -s -w bfs_test true scratch/xattr_link
             ;;
         FreeBSD)
-            setextattr user bfs_test true scratch/xattr
-            setextattr user bfs_test_2 true scratch/xattr_2
-            setextattr -h user bfs_test true scratch/xattr_link
+            setextattr user bfs_test true scratch/xattr \
+                && setextattr user bfs_test_2 true scratch/xattr_2 \
+                && setextattr -h user bfs_test true scratch/xattr_link
             ;;
         *)
             # Linux tmpfs doesn't support the user.* namespace, so we use the security.*
             # namespace, which is writable by root and readable by others
-            sudo setfattr -n security.bfs_test scratch/xattr
-            sudo setfattr -n security.bfs_test_2 scratch/xattr_2
-            sudo setfattr -h -n security.bfs_test scratch/xattr_link
+            sudo setfattr -n security.bfs_test scratch/xattr \
+                && sudo setfattr -n security.bfs_test_2 scratch/xattr_2 \
+                && sudo setfattr -h -n security.bfs_test scratch/xattr_link
             ;;
     esac
 }
 
 function test_xattr() {
-    if ! quiet invoke_bfs scratch -quit -xattr; then
-        return 0
-    fi
-
-    make_xattrs
+    quiet invoke_bfs scratch -quit -xattr || return 0
+    make_xattrs || return 0
     bfs_diff scratch -xattr
 }
 
 function test_L_xattr() {
-    if ! quiet invoke_bfs scratch -quit -xattr; then
-        return 0
-    fi
-
-    make_xattrs
+    quiet invoke_bfs scratch -quit -xattr || return 0
+    make_xattrs || return 0
     bfs_diff -L scratch -xattr
 }
 
 function test_xattrname() {
-    if ! quiet invoke_bfs scratch -quit -xattr; then
-        return 0
-    fi
-
-    make_xattrs
+    quiet invoke_bfs scratch -quit -xattr || return 0
+    make_xattrs || return 0
 
     case "$UNAME" in
         Darwin|FreeBSD)
@@ -2818,11 +2888,8 @@ function test_xattrname() {
 }
 
 function test_L_xattrname() {
-    if ! quiet invoke_bfs scratch -quit -xattr; then
-        return 0
-    fi
-
-    make_xattrs
+    quiet invoke_bfs scratch -quit -xattr || return 0
+    make_xattrs || return 0
 
     case "$UNAME" in
         Darwin|FreeBSD)
@@ -2921,6 +2988,17 @@ function test_exclude_exclude() {
     ! quiet invoke_bfs basic -exclude -exclude -name foo
 }
 
+function test_flags() {
+    quiet invoke_bfs scratch -quit -flags offline || return 0
+
+    rm -rf scratch/*
+
+    touch scratch/{foo,bar}
+    chflags offline scratch/bar || return 0
+
+    bfs_diff scratch -flags -offline,nohidden
+}
+
 
 BOL=
 EOL='\n'
