@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2020 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -19,8 +19,8 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <langinfo.h>
+#include <nl_types.h>
 #include <regex.h>
-#include <stdarg.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -39,23 +39,9 @@
 #	include <sys/mkdev.h>
 #endif
 
-int xreaddir(DIR *dir, struct dirent **de) {
-	while (true) {
-		errno = 0;
-		*de = readdir(dir);
-
-		if (*de) {
-			const char *name = (*de)->d_name;
-			if (name[0] != '.' || (name[1] != '\0' && (name[1] != '.' || name[2] != '\0'))) {
-				return 0;
-			}
-		} else if (errno != 0) {
-			return -1;
-		} else {
-			return 0;
-		}
-	}
-}
+#if BFS_HAS_UTIL
+#	include <util.h>
+#endif
 
 char *xreadlinkat(int fd, const char *path, size_t size) {
 	ssize_t len;
@@ -250,6 +236,32 @@ int xfaccessat(int fd, const char *path, int amode) {
 	return ret;
 }
 
+int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *clear) {
+#if BSD && !__GNU__
+	char *str_arg = (char *)*str;
+	unsigned long set_arg = 0;
+	unsigned long clear_arg = 0;
+
+#if __NetBSD__
+	int ret = string_to_flags(&str_arg, &set_arg, &clear_arg);
+#else
+	int ret = strtofflags(&str_arg, &set_arg, &clear_arg);
+#endif
+
+	*str = str_arg;
+	*set = set_arg;
+	*clear = clear_arg;
+
+	if (ret != 0) {
+		errno = EINVAL;
+	}
+	return ret;
+#else // !BSD
+	errno = ENOTSUP;
+	return -1;
+#endif
+}
+
 bool is_nonexistence_error(int error) {
 	return error == ENOENT || errno == ENOTDIR;
 }
@@ -262,7 +274,7 @@ static char *xgetline(void) {
 	}
 
 	while (true) {
-		int c = getchar();
+		int c = fgetc(stdin);
 		if (c == '\n' || c == EOF) {
 			break;
 		}
@@ -356,3 +368,49 @@ int bfs_minor(dev_t dev) {
 	return dev & 0xFF;
 #endif
 }
+
+size_t xread(int fd, void *buf, size_t nbytes) {
+	size_t count = 0;
+
+	while (count < nbytes) {
+		ssize_t ret = read(fd, (char *)buf + count, nbytes - count);
+		if (ret < 0) {
+			if (errno == EINTR) {
+				continue;
+			} else {
+				break;
+			}
+		} else if (ret == 0) {
+			// EOF
+			errno = 0;
+			break;
+		} else {
+			count += ret;
+		}
+	}
+
+	return count;
+}
+
+size_t xwrite(int fd, const void *buf, size_t nbytes) {
+	size_t count = 0;
+
+	while (count < nbytes) {
+		ssize_t ret = write(fd, (const char *)buf + count, nbytes - count);
+		if (ret < 0) {
+			if (errno == EINTR) {
+				continue;
+			} else {
+				break;
+			}
+		} else if (ret == 0) {
+			// EOF?
+			errno = 0;
+			break;
+		} else {
+			count += ret;
+		}
+	}
+
+	return count;
+}
