@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2020 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2021 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -21,11 +21,11 @@
 #ifndef BFS_UTIL_H
 #define BFS_UTIL_H
 
-#include <dirent.h>
 #include <fcntl.h>
 #include <fnmatch.h>
 #include <regex.h>
 #include <stdbool.h>
+#include <stddef.h>
 #include <sys/types.h>
 
 // Some portability concerns
@@ -42,6 +42,12 @@
 #	define BFS_HAS_INCLUDE(header, fallback) fallback
 #endif
 
+#ifdef __has_c_attribute
+#	define BFS_HAS_C_ATTRIBUTE(attr) __has_c_attribute(attr)
+#else
+#	define BFS_HAS_C_ATTRIBUTE(attr) false
+#endif
+
 #ifndef BFS_HAS_MNTENT
 #	define BFS_HAS_MNTENT BFS_HAS_INCLUDE(<mntent.h>, __GLIBC__)
 #endif
@@ -74,6 +80,10 @@
 #	define BFS_HAS_SYS_XATTR BFS_HAS_INCLUDE(<sys/xattr.h>, __linux__)
 #endif
 
+#ifndef BFS_HAS_UTIL
+#	define BFS_HAS_UTIL BFS_HAS_INCLUDE(<util.h>, __NetBSD__)
+#endif
+
 #if !defined(FNM_CASEFOLD) && defined(FNM_IGNORECASE)
 #	define FNM_CASEFOLD FNM_IGNORECASE
 #endif
@@ -82,6 +92,14 @@
 #	define O_DIRECTORY 0
 #endif
 
+#if BFS_HAS_C_ATTRIBUTE(fallthrough)
+#	define fallthrough [[fallthrough]]
+#elif __GNUC__
+#	define fallthrough __attribute__((fallthrough))
+#else
+#	define fallthrough
+#endif
+
 /**
  * Adds compiler warnings for bad printf()-style function calls, if supported.
  */
@@ -91,10 +109,27 @@
 #	define BFS_FORMATTER(fmt, args)
 #endif
 
+// Lower bound on BFS_FLEX_SIZEOF()
+#define BFS_FLEX_LB(type, member, length) (offsetof(type, member) + sizeof(((type *)NULL)->member[0]) * (length))
+
+// Maximum macro for BFS_FLEX_SIZE()
+#define BFS_FLEX_MAX(a, b) ((a) > (b) ? (a) : (b))
+
 /**
- * readdir() wrapper that makes error handling cleaner.
+ * Computes the size of a struct containing a flexible array member of the given
+ * length.
+ *
+ * @param type
+ *         The type of the struct containing the flexible array.
+ * @param member
+ *         The name of the flexible array member.
+ * @param length
+ *         The length of the flexible array.
  */
-int xreaddir(DIR *dir, struct dirent **de);
+#define BFS_FLEX_SIZEOF(type, member, length) \
+	(sizeof(type) <= BFS_FLEX_LB(type, member, 0) \
+		? BFS_FLEX_LB(type, member, length) \
+		: BFS_FLEX_MAX(sizeof(type), BFS_FLEX_LB(type, member, length)))
 
 /**
  * readlinkat() wrapper that dynamically allocates the result.
@@ -109,26 +144,6 @@ int xreaddir(DIR *dir, struct dirent **de);
  */
 char *xreadlinkat(int fd, const char *path, size_t size);
 
-/**
- * Check if a file descriptor is open.
- */
-bool isopen(int fd);
-
-/**
- * Open a file and redirect it to a particular descriptor.
- *
- * @param fd
- *         The file descriptor to redirect.
- * @param path
- *         The path to open.
- * @param flags
- *         The flags passed to open().
- * @param mode
- *         The mode passed to open() (optional).
- * @return fd on success, -1 on failure.
- */
-int redirect(int fd, const char *path, int flags, ...);
-
 /**
  * Like dup(), but set the FD_CLOEXEC flag.
  *
@@ -166,7 +181,7 @@ char *xregerror(int err, const regex_t *regex);
  * @param str
  *         The string to hold the formatted mode.
  */
-void format_mode(mode_t mode, char str[11]);
+void xstrmode(mode_t mode, char str[11]);
 
 /**
  * basename() variant that doesn't modify the input.
@@ -182,6 +197,21 @@ const char *xbasename(const char *path);
  */
 int xfaccessat(int fd, const char *path, int amode);
 
+/**
+ * Portability wrapper for strtofflags().
+ *
+ * @param str
+ *         The string to parse.  The pointee will be advanced to the first
+ *         invalid position on error.
+ * @param set
+ *         The flags that are set in the string.
+ * @param clear
+ *         The flags that are cleared in the string.
+ * @return
+ *         0 on success, -1 on failure.
+ */
+int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *clear);
+
 /**
  * Return whether an error code is due to a path not existing.
  */
@@ -209,4 +239,23 @@ int bfs_major(dev_t dev);
  */
 int bfs_minor(dev_t dev);
 
+/**
+ * A safe version of read() that handles interrupted system calls and partial
+ * reads.
+ *
+ * @return
+ *         The number of bytes read.  A value != nbytes indicates an error
+ *         (errno != 0) or end of file (errno == 0).
+ */
+size_t xread(int fd, void *buf, size_t nbytes);
+
+/**
+ * A safe version of write() that handles interrupted system calls and partial
+ * writes.
+ *
+ * @return
+           The number of bytes written.  A value != nbytes indicates an error.
+ */
+size_t xwrite(int fd, const void *buf, size_t nbytes);
+
 #endif // BFS_UTIL_H
